from ewokscore import execute_graph
from ewokscore import Task
from ewoksutils.import_utils import qualname


class SumTask(
    Task,
    input_names=["a"],
    optional_input_names=["b"],
    output_names=["result", "inputs", "label"],
):
    def run(self):
        result = self.inputs.a
        if self.inputs.b:
            result += self.inputs.b
        self.outputs.result = result
        self.outputs.inputs = {k: v for k, v in self.get_input_values().items() if v}
        self.outputs.label = self.label


def create_graph():
    task = qualname(SumTask)
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task2",
            "default_inputs": [{"name": "a", "value": 2}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task3",
            "default_inputs": [{"name": "b", "value": 3}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task4",
            "default_inputs": [{"name": "b", "value": 4}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task5",
            "default_inputs": [{"name": "b", "value": 5}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task6",
            "default_inputs": [{"name": "b", "value": 6}],
            "task_type": "class",
            "task_identifier": task,
        },
        {
            "id": "task7",
            "task_type": "class",
            "task_identifier": "ewokscore.tests.examples.tasks.nooutputtask.NoOutputTask",
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task4",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task3",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task4",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "b"}],
        },
        {
            "source": "task5",
            "target": "task6",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task6",
            "target": "task7",
            "map_all_data": True,
        },
    ]

    return {
        "graph": {"id": "testgraph"},
        "links": links,
        "nodes": nodes,
    }


def test_execute_graph_outputs():
    # All task instances
    results = execute_graph(create_graph(), output_tasks=True)
    results = {k: v.get_output_values() for k, v in results.items()}
    expected = {
        "task1": {"inputs": {"a": 1}, "result": 1, "label": "task1"},
        "task3": {"inputs": {"a": 1, "b": 3}, "result": 4, "label": "task3"},
        "task2": {"inputs": {"a": 2}, "result": 2, "label": "task2"},
        "task4": {"inputs": {"a": 2, "b": 4}, "result": 6, "label": "task4"},
        "task5": {"inputs": {"a": 4, "b": 6}, "result": 10, "label": "task5"},
        "task6": {"inputs": {"a": 10, "b": 6}, "result": 16, "label": "task6"},
        "task7": {},
    }
    assert results == expected

    # The results of all tasks
    results = execute_graph(
        create_graph(), outputs=[{"all": True}], merge_outputs=False
    )
    assert results == expected

    # Merge the results of all tasks
    results = execute_graph(create_graph(), outputs=[{"all": True}])
    expected = {"inputs": {"a": 10, "b": 6}, "result": 16, "label": "task6"}
    assert results == expected

    # The results of selected tasks
    results = execute_graph(
        create_graph(), outputs=[{"id": "task5"}], merge_outputs=False
    )
    expected = {"task5": {"inputs": {"a": 4, "b": 6}, "result": 10, "label": "task5"}}
    assert results == expected

    # Merge the results of selected tasks
    results = execute_graph(create_graph(), outputs=[{"id": "task5"}])
    expected = {"inputs": {"a": 4, "b": 6}, "result": 10, "label": "task5"}
    assert results == expected

    # The results of selected tasks
    results = execute_graph(
        create_graph(),
        outputs=[
            {"id": "task1", "name": "inputs"},
            {"id": "task4", "name": "result"},
        ],
        merge_outputs=False,
    )
    expected = {"task1": {"inputs": {"a": 1}}, "task4": {"result": 6}}
    assert results == expected

    # Merge the results of selected tasks
    results = execute_graph(
        create_graph(),
        outputs=[
            {"id": "task1", "name": "inputs"},
            {"id": "task4", "name": "result"},
        ],
    )
    expected = {"inputs": {"a": 1}, "result": 6}
    assert results == expected

    # The results of selected tasks
    results = execute_graph(
        create_graph(),
        outputs=[
            {"id": "task1", "name": "inputs", "new_name": "a"},
            {"id": "task4", "name": "result"},
        ],
        merge_outputs=False,
    )
    expected = {"task1": {"a": {"a": 1}}, "task4": {"result": 6}}
    assert results == expected

    # Merge the results of selected tasks
    results = execute_graph(
        create_graph(),
        outputs=[
            {"id": "task1", "name": "inputs", "new_name": "a"},
            {"id": "task4", "name": "result"},
        ],
    )
    expected = {"a": {"a": 1}, "result": 6}
    assert results == expected


def test_execute_graph_inputs():
    results = execute_graph(
        create_graph(),
        inputs=[{"id": "task1", "name": "b", "value": 1}],
        output_tasks=True,
    )
    results = {k: v.get_output_values() for k, v in results.items()}
    expected = {
        "task1": {"inputs": {"a": 1, "b": 1}, "result": 2, "label": "task1"},
        "task3": {"inputs": {"a": 2, "b": 3}, "result": 5, "label": "task3"},
        "task2": {"inputs": {"a": 2}, "result": 2, "label": "task2"},
        "task4": {"inputs": {"a": 2, "b": 4}, "result": 6, "label": "task4"},
        "task5": {"inputs": {"a": 5, "b": 6}, "result": 11, "label": "task5"},
        "task6": {"inputs": {"a": 11, "b": 6}, "result": 17, "label": "task6"},
        "task7": {},
    }
    assert results == expected

    results = execute_graph(
        create_graph(),
        inputs=[{"name": "b", "value": 1}],
        output_tasks=True,
    )
    results = {k: v.get_output_values() for k, v in results.items()}
    expected = {
        "task1": {"inputs": {"a": 1, "b": 1}, "result": 2, "label": "task1"},
        "task3": {"inputs": {"a": 2, "b": 3}, "result": 5, "label": "task3"},
        "task2": {"inputs": {"a": 2, "b": 1}, "result": 3, "label": "task2"},
        "task4": {"inputs": {"a": 3, "b": 4}, "result": 7, "label": "task4"},
        "task5": {"inputs": {"a": 5, "b": 7}, "result": 12, "label": "task5"},
        "task6": {"inputs": {"a": 12, "b": 6}, "result": 18, "label": "task6"},
        "task7": {},
    }
    assert results == expected



import pytest
from ewokscore.graph import load_graph
from ewokscore.graph.analysis import link_is_required


def test_required_links():
    nodes = [
        {"id": "source1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "source2a", "task_type": "method", "task_identifier": "dummy"},
        {"id": "source2b", "task_type": "method", "task_identifier": "dummy"},
        {"id": "target", "task_type": "method", "task_identifier": "dummy"},
    ]
    links = [
        {"source": "source1", "target": "target"},
        {"source": "source2a", "target": "source2b"},
        {"source": "source2b", "target": "target"},
    ]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})
    assert link_is_required(taskgraph.graph, "source1", "target")
    assert link_is_required(taskgraph.graph, "source2a", "source2b")
    assert link_is_required(taskgraph.graph, "source2b", "target")

    links[0]["conditions"] = [{"source_output": "a", "value": 1}]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})
    assert not link_is_required(taskgraph.graph, "source1", "target")
    assert link_is_required(taskgraph.graph, "source2a", "source2b")
    assert link_is_required(taskgraph.graph, "source2b", "target")
    links[0].pop("conditions")

    links[1]["conditions"] = [{"source_output": "a", "value": 1}]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})
    assert link_is_required(taskgraph.graph, "source1", "target")
    assert not link_is_required(taskgraph.graph, "source2a", "source2b")
    assert not link_is_required(taskgraph.graph, "source2b", "target")
    links[1].pop("conditions")

    links[1]["conditions"] = [{"source_output": "a", "value": 1}]
    links[1]["required"] = True
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})
    assert link_is_required(taskgraph.graph, "source1", "target")
    assert link_is_required(taskgraph.graph, "source2a", "source2b")
    assert link_is_required(taskgraph.graph, "source2b", "target")
    links[1].pop("conditions")

    links[2]["conditions"] = [{"source_output": "a", "value": 1}]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})
    assert link_is_required(taskgraph.graph, "source1", "target")
    assert link_is_required(taskgraph.graph, "source2a", "source2b")
    assert not link_is_required(taskgraph.graph, "source2b", "target")
    links[2].pop("conditions")

    links[2]["conditions"] = [{"source_output": "a", "value": 1}]
    links[2]["required"] = True
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})
    assert link_is_required(taskgraph.graph, "source1", "target")
    assert link_is_required(taskgraph.graph, "source2a", "source2b")
    assert link_is_required(taskgraph.graph, "source2b", "target")
    links[2].pop("conditions")


def test_wrong_argument_definitions():
    nodes = [
        {"id": "source1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "source2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "target", "task_type": "method", "task_identifier": "dummy"},
    ]
    links = [
        {
            "source": "source1",
            "target": "target",
            "data_mapping": [{"source_output": "a", "target_input": "a"}],
        },
        {
            "source": "source2",
            "target": "target",
            "data_mapping": [{"source_output": "a", "target_input": "a"}],
        },
    ]
    graph = {"graph": {"id": "test"}, "nodes": nodes, "links": links}
    with pytest.raises(ValueError):
        load_graph(graph)

    links[0]["conditions"] = [{"source_output": "a", "value": 1}]
    load_graph(graph)

    links[0]["required"] = True
    with pytest.raises(ValueError):
        load_graph(graph)



from typing import Any, Optional, List, Union
from .graph import load_graph as _load_graph
from .graph import TaskGraph
from .graph.execute import sequential
from .graph.graph_io import update_default_inputs
from .events import job_decorator as execute_graph_decorator

__all__ = [
    "execute_graph",
    "load_graph",
    "save_graph",
    "convert_graph",
    "graph_is_supported",
    "execute_graph_decorator",
]


def load_graph(
    graph: Any, inputs: Optional[List[dict]] = None, **load_options
) -> TaskGraph:
    taskgraph = _load_graph(source=graph, **load_options)
    if inputs:
        update_default_inputs(taskgraph.graph, inputs)
    return taskgraph


def save_graph(graph: TaskGraph, destination, **save_options) -> Union[str, dict]:
    return graph.dump(destination, **save_options)


def convert_graph(
    source,
    destination,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    save_options: Optional[dict] = None,
) -> Union[str, dict]:
    if load_options is None:
        load_options = dict()
    if save_options is None:
        save_options = dict()
    graph = load_graph(source, inputs=inputs, **load_options)
    return save_graph(graph, destination, **save_options)


@execute_graph_decorator()
def execute_graph(
    graph,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    **execute_options
):
    if load_options is None:
        load_options = dict()
    taskgraph = load_graph(graph, inputs=inputs, **load_options)
    return sequential.execute_graph(taskgraph.graph, **execute_options)


def graph_is_supported(graph: TaskGraph) -> bool:
    return not graph.is_cyclic and not graph.has_conditional_links



from ewoksutils.import_utils import qualname
from ewokscore.task import Task


def mymethod1(a=0, b=0):
    return a + b


def test_method_task1(varinfo):
    task = Task.instantiate(
        "MethodExecutorTask",
        inputs={"_method": qualname(mymethod1), "a": 3, "b": 5},
        varinfo=varinfo,
    )
    task.execute()
    assert task.done
    assert task.get_output_values() == {"return_value": 8}


def mymethod2(*args):
    return sum(args)


def test_method_task2(varinfo):
    task = Task.instantiate(
        "MethodExecutorTask",
        inputs={"_method": qualname(mymethod2), 0: 3, 1: 5},
        varinfo=varinfo,
    )
    task.execute()
    assert task.done
    assert task.get_output_values() == {"return_value": 8}


def mymethod3(a, *args, b=None, c=3, **kw):
    print("a:", a)
    print("b:", b)
    print("c:", c)
    print("args:", args)
    print("kwargs:", kw)
    return a + sum(args) + b + c + sum(kw.values())


def test_method_task3(varinfo):
    task = Task.instantiate(
        "MethodExecutorTask",
        inputs={"_method": qualname(mymethod3), 0: 2, 1: 4, "b": 7, "d": 10},
        varinfo=varinfo,
    )
    task.execute()
    assert task.done
    assert task.get_output_values() == {"return_value": 26}


def myppfmethod(a=0, b=0, **kw):
    return {"a": a + b}


def test_ppfmethod_task(varinfo):
    task = Task.instantiate(
        "PpfMethodExecutorTask",
        inputs={"_method": qualname(myppfmethod), "a": 3, "b": 5},
        varinfo=varinfo,
    )
    task.execute()
    assert task.done
    assert task.get_output_values() == {"_ppfdict": {"a": 8, "b": 5}}



from collections.abc import Mapping, Sequence


def dict_merge(
    destination, source, overwrite=False, _nodes=None, contatenate_sequences=False
):
    """Merge the source into the destination"""
    if _nodes is None:
        _nodes = tuple()
    for key, value in source.items():
        if key in destination:
            _nodes += (str(key),)
            if isinstance(destination[key], Mapping) and isinstance(value, Mapping):
                dict_merge(
                    destination[key],
                    value,
                    overwrite=overwrite,
                    _nodes=_nodes,
                    contatenate_sequences=contatenate_sequences,
                )
            elif value == destination[key]:
                continue
            elif overwrite:
                destination[key] = value
            elif (
                contatenate_sequences
                and isinstance(destination[key], Sequence)
                and isinstance(value, Sequence)
            ):
                destination[key] += value
            else:
                raise ValueError("Conflict at " + ".".join(_nodes))
        else:
            destination[key] = value



"""Ewoks sends job, workflow and node events during workflow execution.
"""

from .global_state import add_handler  # noqa F401
from .global_state import remove_handler  # noqa F401
from .global_state import cleanup  # noqa F401
from .send_events import send_job_event  # noqa F401
from .send_events import send_workflow_event  # noqa F401
from .send_events import send_task_event  # noqa F401
from .contexts import job_decorator  # noqa F401
from .contexts import job_context  # noqa F401
from .contexts import workflow_context  # noqa F401
from .contexts import node_context  # noqa F401



import warnings
from typing import Optional, Union, Mapping

from .hashing import UniversalHashable
from .registration import Registered
from .variable import VariableContainer
from .variable import VariableContainerNamespace
from .variable import ReadOnlyVariableContainerNamespace
from .variable import VariableContainerMissingNamespace
from . import node
from . import missing_data
from . import events


class TaskInputError(ValueError):
    pass


class Task(Registered, UniversalHashable, register=False):
    """Node in a task Graph with named inputs and outputs.

    The universal hash of the task is equal to the universal
    hash of the output. The universal hash of the output is
    equal to the hash of the inputs and the task nonce.

    A task is done when its output exists.

    This is an abstract class. Instantiating a `Task` should be
    done with `ewokscore.inittask.instantiate_task`.
    """

    _INPUT_NAMES = set()
    _OPTIONAL_INPUT_NAMES = set()
    _OUTPUT_NAMES = set()
    _N_REQUIRED_POSITIONAL_INPUTS = 0

    def __init__(
        self,
        inputs: Optional[Mapping] = None,
        varinfo: Optional[dict] = None,
        node_id: Optional[node.NodeIdType] = None,
        node_attrs: Optional[dict] = None,
        execinfo: Optional[dict] = None,
    ):
        """The named arguments are inputs and Variable configuration"""
        if inputs is None:
            inputs = dict()
        elif not isinstance(inputs, Mapping):
            raise TypeError(inputs, type(inputs))

        # Check required inputs
        missing_required = set(self._INPUT_NAMES) - set(inputs.keys())
        if missing_required:
            raise TaskInputError(f"Missing inputs for {type(self)}: {missing_required}")

        # Check required positional inputs
        nrequiredargs = self._N_REQUIRED_POSITIONAL_INPUTS
        for i in range(nrequiredargs):
            if i not in inputs and str(i) not in inputs:
                raise TaskInputError(
                    f"Missing inputs for {type(self)}: positional argument #{i}"
                )

        # Init missing optional inputs
        missing_optional = set(self._OPTIONAL_INPUT_NAMES) - set(inputs.keys())
        for varname in missing_optional:
            inputs[varname] = self.MISSING_DATA

        # Required outputs for the task to be "done"
        ovars = {varname: self.MISSING_DATA for varname in self._OUTPUT_NAMES}

        # Node/task info
        node_id = node.get_node_id(node_id, node_attrs)
        self.__node_id = node_id
        self.__node_label = node.get_node_label(node_id, node_attrs)
        task_id = self.class_registry_name()
        task_id = node.get_task_identifier(node_attrs, task_id)
        self.__task_id = task_id
        if node_id and task_id:
            self.__execinfo = execinfo
        else:
            self.__execinfo = None

        # Misc
        self.__exception = None
        self.__succeeded = None

        # The output hash will update dynamically if any of the input
        # variables change
        varinfo = node.get_varinfo(node_attrs, varinfo)
        self.__inputs = VariableContainer(value=inputs, varinfo=varinfo)
        self.__outputs = VariableContainer(
            value=ovars,
            pre_uhash=self.__inputs,
            instance_nonce=self.class_nonce(),
            varinfo=varinfo,
        )

        self.__inputs_namespace = ReadOnlyVariableContainerNamespace(self.__inputs)
        self.__outputs_namespace = VariableContainerNamespace(self.__outputs)

        self.__missing_inputs_namespace = VariableContainerMissingNamespace(
            self.__inputs
        )
        self.__missing_outputs_namespace = VariableContainerMissingNamespace(
            self.__outputs
        )

        # The task class has the same hash as its output
        super().__init__(pre_uhash=self.__outputs)

    def __init_subclass__(
        subclass,
        input_names=tuple(),
        optional_input_names=tuple(),
        output_names=tuple(),
        n_required_positional_inputs=0,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)
        input_names = set(input_names)
        optional_input_names = set(optional_input_names)
        output_names = set(output_names)

        reserved = subclass._reserved_variable_names()
        forbidden = input_names & reserved
        forbidden |= optional_input_names & reserved
        forbidden |= output_names & reserved
        if forbidden:
            raise RuntimeError(
                "The following names cannot be used a variable names: "
                + str(list(forbidden))
            )

        # Ensures that each subclass has their own sets:
        subclass._INPUT_NAMES = subclass._INPUT_NAMES | set(input_names)
        subclass._OPTIONAL_INPUT_NAMES = subclass._OPTIONAL_INPUT_NAMES | set(
            optional_input_names
        )
        subclass._OUTPUT_NAMES = subclass._OUTPUT_NAMES | set(output_names)
        subclass._N_REQUIRED_POSITIONAL_INPUTS = n_required_positional_inputs

    @staticmethod
    def _reserved_variable_names():
        return VariableContainerNamespace._reserved_variable_names()

    @classmethod
    def instantiate(cls, registry_name: str, **kw):
        r"""Factory method for instantiating a derived class.

        :param str registry_name: for example "tasklib.tasks.MyTask" or "MyTask"
        :param \**kw: `Task` constructor arguments
        :returns Task:
        """
        return cls.get_subclass(registry_name)(**kw)

    @classmethod
    def required_input_names(cls):
        return cls._INPUT_NAMES

    @classmethod
    def optional_input_names(cls):
        return cls._OPTIONAL_INPUT_NAMES

    @classmethod
    def input_names(cls):
        return cls._INPUT_NAMES | cls._OPTIONAL_INPUT_NAMES

    @classmethod
    def output_names(cls):
        return cls._OUTPUT_NAMES

    @classmethod
    def class_nonce_data(cls):
        return super().class_nonce_data() + (
            sorted(cls.input_names()),
            sorted(cls.output_names()),
            cls._N_REQUIRED_POSITIONAL_INPUTS,
        )

    @property
    def input_variables(self):
        if self.__inputs is None:
            raise RuntimeError("references have been removed")
        return self.__inputs

    @property
    def inputs(self):
        return self.__inputs_namespace

    @property
    def missing_inputs(self):
        return self.__missing_inputs_namespace

    def get_input_value(self, key, default=missing_data.MISSING_DATA):
        if self.missing_inputs[key]:
            return default
        return self.inputs[key]

    @property
    def input_uhashes(self):
        return self.get_input_uhashes()

    def get_input_uhashes(self):
        return self.__inputs.get_variable_uhashes()

    @property
    def input_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'input_values' is deprecated in favor of the function 'get_input_values'",
            DeprecationWarning,
        )
        return self.get_input_values()

    def get_input_values(self):
        return self.__inputs.get_variable_values()

    @property
    def named_input_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'named_input_values' is deprecated in favor of the function 'get_named_input_values'",
            DeprecationWarning,
        )
        return self.get_named_input_values()

    def get_named_input_values(self):
        return self.__inputs.get_named_variable_values()

    @property
    def positional_input_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'positional_input_values' is deprecated in favor of the function 'get_positional_input_values'",
            DeprecationWarning,
        )
        return self.__inputs.get_positional_input_values()

    def get_positional_input_values(self):
        return self.__inputs.get_positional_variable_values()

    @property
    def npositional_inputs(self):
        return self.__inputs.n_positional_variables

    @property
    def output_variables(self):
        return self.__outputs

    @property
    def missing_outputs(self):
        return self.__missing_outputs_namespace

    @property
    def outputs(self):
        return self.__outputs_namespace

    def get_output_value(self, key, default=missing_data.MISSING_DATA):
        if self.missing_outputs[key]:
            return default
        return self.outputs[key]

    @property
    def output_uhashes(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'output_uhashes' is deprecated in favor of the function 'get_output_uhashes'",
            DeprecationWarning,
        )
        return self.get_output_uhashes()

    def get_output_uhashes(self):
        return self.__outputs.get_variable_uhashes()

    @property
    def output_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'output_values' is deprecated in favor of the function 'get_output_values'",
            DeprecationWarning,
        )
        return self.get_output_values()

    def get_output_values(self):
        return self.__outputs.get_variable_values()

    @property
    def output_transfer_data(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'output_transfer_data' is deprecated in favor of the function 'get_output_transfer_data'",
            DeprecationWarning,
        )
        return self.get_output_transfer_data()

    def get_output_transfer_data(self):
        """The values are either `DataUri` or `Variable`"""
        return self.__outputs.get_variable_transfer_data()

    @property
    def output_metadata(self) -> Union[dict, None]:
        return self.__outputs.metadata

    def _update_output_metadata(self):
        metadata = self.output_metadata
        if metadata is None:
            return
        if self.__node_label:
            metadata.setdefault("title", self.__node_label)

    @property
    def done(self):
        """Completed (with or without exception)"""
        return self.failed or self.succeeded

    @property
    def succeeded(self):
        """Completed without exception and with output values"""
        if self._OUTPUT_NAMES:
            return self.__outputs.has_value
        else:
            return self.__succeeded

    @property
    def failed(self):
        """Completed with exception"""
        return self.__exception is not None

    @property
    def exception(self):
        return self.__exception

    def _get_repr_data(self):
        data = super()._get_repr_data()
        if self.__node_label:
            data["label"] = repr(str(self.__node_label))
        else:
            data["label"] = None

    @property
    def label(self):
        if self.__node_label:
            return self.__node_label
        else:
            return str(self)

    @property
    def node_id(self):
        return self.__node_id

    @property
    def job_id(self) -> Optional[str]:
        if self.__execinfo:
            return self.__execinfo.get("job_id")

    @property
    def workflow_id(self) -> Optional[str]:
        if self.__execinfo:
            return self.__execinfo.get("workflow_id")

    def _iter_missing_input_values(self):
        for iname in self._INPUT_NAMES:
            var = self.__inputs.get(iname)
            if var is None or not var.has_value:
                yield iname

    @property
    def is_ready_to_execute(self):
        try:
            next(iter(self._iter_missing_input_values()))
        except StopIteration:
            return True
        return False

    def assert_ready_to_execute(self):
        lst = list(self._iter_missing_input_values())
        if lst:
            raise TaskInputError(
                "The following inputs could not be loaded: " + str(lst)
            )

    def execute(
        self,
        force_rerun: Optional[bool] = False,
        raise_on_error: Optional[bool] = True,
        cleanup_references: Optional[bool] = False,
    ):
        with events.node_context(
            self.__execinfo, node_id=self.__node_id, task_id=self.__task_id
        ) as execinfo:
            self.__execinfo = execinfo
            self._send_start_event()
            try:
                if force_rerun:
                    # Rerun a task which is already done
                    self.__outputs.force_non_existing()
                if self.done:
                    return
                self.assert_ready_to_execute()
                self.run()
                self._update_output_metadata()
                self.__outputs.dump()
            except Exception as e:
                self.__exception = e
                if raise_on_error:
                    raise RuntimeError(f"Task '{self.label}' failed") from e
            else:
                self.__succeeded = True
            finally:
                if cleanup_references:
                    self.cleanup_references()
                self._send_send_event()

    def _send_event(self, **kwargs):
        """Send an ewoks event"""
        if self.__execinfo:
            events.send_task_event(execinfo=self.__execinfo, **kwargs)

    def _send_start_event(self):
        input_uris = [
            {"name": name, "value": str(uri) if uri else None}
            for name, uri in self.__inputs.get_variable_uris().items()
        ]
        output_uris = [
            {"name": name, "value": str(uri) if uri else None}
            for name, uri in self.__outputs.get_variable_uris().items()
        ]
        task_uri = self.__outputs.data_uri
        if task_uri:
            task_uri = str(task_uri)
        self._send_event(
            event="start",
            input_uris=input_uris,
            output_uris=output_uris,
            task_uri=task_uri,
        )

    def _send_send_event(self):
        self._send_event(event="end", exception=self.exception)

    def cleanup_references(self):
        """Removes all references to the inputs.
        Side effect: fixes the uhash of the task and outputs
        """
        self.__inputs = None
        self.__inputs_namespace = None
        self.__missing_inputs_namespace = None
        self.__outputs.cleanup_references()
        super().cleanup_references()

    def run(self):
        """To be implemented by the derived classes"""
        raise NotImplementedError



import sys
import pkgutil
import inspect
import logging
from fnmatch import fnmatch
from types import ModuleType
from typing import Callable, Iterable, Generator, Optional, List, Dict, Tuple

if sys.version_info < (3, 9):
    from importlib_metadata import entry_points as _entry_points

    def iter_entry_points(group: str):
        return _entry_points(group=group)

elif sys.version_info < (3, 10):
    from importlib.metadata import entry_points as _entry_points

    def iter_entry_points(group: str):
        return _entry_points().get(group, [])

else:
    from importlib.metadata import entry_points as _entry_points

    def iter_entry_points(group: str):
        return _entry_points(group=group)


from ewoksutils.import_utils import qualname
from ewoksutils.import_utils import import_module

from .task import Task


logger = logging.getLogger(__name__)


def discover_tasks_from_modules(
    *module_names: Iterable[str],
    task_type="class",
    reload: bool = False,
    raise_import_failure: bool = True,
) -> List[Dict[str, str]]:
    return list(
        iter_discover_tasks_from_modules(
            *module_names,
            task_type=task_type,
            reload=reload,
            raise_import_failure=raise_import_failure,
        )
    )


def iter_discover_tasks_from_modules(
    *module_names: Iterable[str],
    task_type="class",
    reload: bool = False,
    raise_import_failure: bool = True,
) -> Generator[Dict[str, str], None, None]:
    if "" not in sys.path:
        # This happens when the python process was launched
        # through a python console script
        sys.path.append("")

    if task_type == "method":
        yield from _iter_method_tasks(
            *module_names, reload=reload, raise_import_failure=raise_import_failure
        )
    elif task_type == "ppfmethod":
        yield from _iter_method_tasks(
            *module_names,
            filter_method_name=lambda name: name == "run",
            reload=reload,
            raise_import_failure=raise_import_failure,
        )
    elif task_type == "class":
        for module_name in module_names:
            _safe_import_module(
                module_name, reload=reload, raise_import_failure=raise_import_failure
            )
        yield from _iter_registered_tasks(*module_names)
    else:
        raise ValueError("Class type does not support discovery")


def _iter_registered_tasks(
    *filter_modules: Iterable[str],
) -> Generator[Dict[str, str], None, None]:
    """Yields all task classes registered in the current process."""
    for cls in Task.get_subclasses():
        module = cls.__module__
        if filter_modules and not any(
            module.startswith(prefix) for prefix in filter_modules
        ):
            continue
        task_identifier = cls.class_registry_name()
        category = task_identifier.split(".")[0]
        yield {
            "task_type": "class",
            "task_identifier": task_identifier,
            "required_input_names": list(cls.required_input_names()),
            "optional_input_names": list(cls.optional_input_names()),
            "output_names": list(cls.output_names()),
            "category": category,
            "description": cls.__doc__,
        }


def _iter_method_tasks(
    *module_names: Iterable[str],
    filter_method_name: Optional[Callable[[str], bool]] = None,
    reload: bool = False,
    raise_import_failure: bool = False,
) -> Generator[Dict[str, str], None, None]:
    """Yields all task methods from the provided module_names. The module_names will be will
    imported for discovery.
    """
    for module_name in module_names:
        mod = _safe_import_module(
            module_name, reload=reload, raise_import_failure=raise_import_failure
        )
        if mod is None:
            continue
        for method in inspect.getmembers(mod, inspect.isfunction):
            method_name, method_qn = method
            if filter_method_name and not filter_method_name(method_name):
                continue
            if method_name.startswith("_"):
                continue
            task_identifier = qualname(method_qn)
            category = task_identifier.split(".")[0]

            method = getattr(mod, method_name)
            required_input_names, optional_input_names = _method_arguments(method)

            yield {
                "task_type": "method",
                "task_identifier": qualname(method_qn),
                "required_input_names": required_input_names,
                "optional_input_names": optional_input_names,
                "output_names": ["return_value"],
                "category": category,
                "description": method.__doc__,
            }


def iter_discover_all_tasks(
    reload: bool = False, raise_import_failure: bool = False
) -> Generator[Dict[str, str], None, None]:
    visited = set()
    for task_type in ("class", "ppfmethod", "method"):
        group = "ewoks.tasks." + task_type
        for entrypoint in iter_entry_points(group):
            module_pattern = entrypoint.name
            if module_pattern is visited:
                continue
            visited.add(module_pattern)
            for module_name in _iter_modules_from_pattern(
                module_pattern, reload=reload, raise_import_failure=raise_import_failure
            ):
                yield from iter_discover_tasks_from_modules(
                    module_name,
                    task_type=task_type,
                    reload=reload,
                    raise_import_failure=raise_import_failure,
                )


def discover_all_tasks(
    reload: bool = False, raise_import_failure: bool = False
) -> Generator[Dict[str, str], None, None]:
    return list(
        iter_discover_all_tasks(
            reload=reload, raise_import_failure=raise_import_failure
        )
    )


def _iter_modules_from_pattern(
    module_pattern: str, reload: bool = False, raise_import_failure: bool = False
) -> Generator[str, None, None]:
    if "*" not in module_pattern:
        yield module_pattern
        return
    ndots = module_pattern.count(".")
    parts = module_pattern.split(".")
    pkg = _safe_import_module(
        parts[0], reload=reload, raise_import_failure=raise_import_failure
    )
    if pkg is None:
        return
    if raise_import_failure:

        def onerror(module_name):
            raise

    else:
        onerror = _onerror
    for pkginfo in pkgutil.walk_packages(
        pkg.__path__, pkg.__name__ + ".", onerror=onerror
    ):
        if pkginfo.name.count(".") == ndots and fnmatch(pkginfo.name, module_pattern):
            yield pkginfo.name


def _safe_import_module(
    module_name: str, reload: bool = False, raise_import_failure: bool = False
) -> Optional[ModuleType]:
    try:
        return import_module(module_name, reload=reload)
    except Exception as e:
        if raise_import_failure:
            raise
        _onerror(module_name, exception=e)


def _onerror(module_name, exception: Exception = None):
    if exception is None:
        exception = sys.exc_info()[1]
    logger.error(f"Module '{module_name}' cannot be imported: {exception}")


def _method_arguments(method) -> Tuple[List[str], List[str]]:
    sig = inspect.signature(method)
    required_input_names = list()
    optional_input_names = list()
    for name, param in sig.parameters.items():
        required = param.default is inspect._empty
        if param.kind == param.VAR_POSITIONAL:
            continue
        if param.kind == param.VAR_KEYWORD:
            continue
        if required:
            required_input_names.append(name)
        else:
            optional_input_names.append(name)
    return required_input_names, optional_input_names



from typing import Any, Dict, Optional, Union
from ewokscore.graph import load_graph
from ewokscore.graph.taskgraph import TaskGraph
from ewokscore.node import NodeIdType
from ewokscore.task import Task
from ewokscore import missing_data
from ewokscore.variable import value_from_transfer
from ewokscore.graph.analysis import end_nodes
from ewokscore.graph.execute.sequential import instantiate_task_static


def assert_execute_graph_default_result(
    taskgraph: Union[TaskGraph, dict],
    result: Dict[str, Any],
    expected: Dict[str, Any],
    varinfo: Optional[dict] = None,
):
    """The default result is the merged output of all end nodes."""
    if varinfo:
        scheme = varinfo.get("scheme")
    else:
        scheme = None
    taskgraph = load_graph(taskgraph)
    if taskgraph.is_cyclic:
        expected = filter_expected_results(
            taskgraph, expected, end_only=True, merge=True
        )
        assert_execute_graph_values(result, expected, varinfo)
    elif scheme:
        _assert_execute_graph_tasks(taskgraph, expected, varinfo=varinfo)
        expected = filter_expected_results(
            taskgraph, expected, end_only=True, merge=True
        )
        assert_execute_graph_values(result, expected, varinfo)
    else:
        expected = filter_expected_results(
            taskgraph, expected, end_only=True, merge=True
        )
        assert_execute_graph_values(result, expected, varinfo)


def assert_execute_graph_tasks(
    taskgraph: Union[TaskGraph, dict],
    result: Dict[str, Any],
    expected: Dict[str, Any],
    varinfo: Optional[dict] = None,
):
    if varinfo:
        scheme = varinfo.get("scheme")
    else:
        scheme = None
    _assert_execute_graph_tasks(
        taskgraph, expected, varinfo=varinfo, execute_graph_result=result
    )
    if scheme and result:
        _assert_execute_graph_tasks(taskgraph, expected, varinfo=varinfo)


def _assert_execute_graph_tasks(
    taskgraph: Union[TaskGraph, dict],
    expected: Dict[NodeIdType, Any],
    varinfo: Optional[dict] = None,
    execute_graph_result: Optional[Dict[NodeIdType, Task]] = None,
):
    """Check the output of `execute_graph` for each node. When a task is not in `execute_graph_result`,
    it will be instantiated.

    An expected value can be:
        * `None`: task is not executed and therefore does not appear in the results
        * `MISSING_DATA`: task has no output and is therefore cannot be persisted
        * else: task is executed and has output
    """
    taskgraph = load_graph(taskgraph)
    assert not taskgraph.is_cyclic, "Can only check DAG results"

    if execute_graph_result is None:
        execute_graph_result = dict()

    for node in taskgraph.graph.nodes:
        task = execute_graph_result.get(node, None)
        loaded = False
        if task is None:
            assert varinfo, "Need 'varinfo' to load task output"
            task = instantiate_task_static(
                taskgraph.graph, node, tasks=execute_graph_result, varinfo=varinfo
            )
            loaded = True
        assert_task_result(task, node, expected, loaded)


def assert_execute_graph_values(
    execute_graph_result: Dict[str, Any],
    expected: Dict[str, Any],
    varinfo: Optional[dict] = None,
):
    """Check the output of `execute_graph` for the selected outputs of the selected nodes."""
    for output_name, expected_value in expected.items():
        value = execute_graph_result[output_name]
        assert_result(value, expected_value, varinfo=varinfo)


def assert_task_result(task: Task, node_id: NodeIdType, expected: dict, loaded: bool):
    expected_value = expected.get(node_id)
    if missing_data.is_missing_data(expected_value):
        if loaded:
            expected_value = None
        else:
            expected_value = dict()

    if expected_value is None:
        assert not task.done, node_id
    else:
        assert task.done, node_id
        try:
            assert task.get_output_values() == expected_value, node_id
        except AssertionError:
            raise
        except Exception as e:
            raise RuntimeError(f"{node_id} does not have a result") from e


def assert_result(value, expected_value, varinfo: Optional[dict] = None):
    value = value_from_transfer(value, varinfo=varinfo)
    assert value == expected_value


def filter_expected_results(
    taskgraph: Union[TaskGraph, dict],
    results: Dict[NodeIdType, Any],
    end_only: bool = False,
    merge: bool = False,
) -> dict:
    taskgraph = load_graph(taskgraph)
    if end_only:
        nodes = end_nodes(taskgraph.graph)
        results = {k: v for k, v in results.items() if k in nodes}
    else:
        nodes = taskgraph.nodes()
    if merge:
        ret = dict()
        for node_id in nodes:
            adict = results.get(node_id)
            if adict:
                ret.update(adict)
        results = ret
    return results



from ewokscore.graph import load_graph
from ewokscore.graph.analysis import link_is_required


def test_graph_link_is_required_conditions1():
    nodes = [
        {"id": "start", "task_type": "method", "task_identifier": "dummy"},
        {"id": "fan", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_false1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_false2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "merge", "task_type": "method", "task_identifier": "dummy"},
        {"id": "end", "task_type": "method", "task_identifier": "dummy"},
    ]
    links = [
        {"source": "start", "target": "fan", "map_all_data": True},
        {
            "source": "fan",
            "target": "on_false1",
            "map_all_data": True,
            "conditions": [{"source_output": "result", "value": False}],
        },
        {
            "source": "fan",
            "target": "on_true1",
            "map_all_data": True,
            "conditions": [{"source_output": "result", "value": True}],
        },
        {"source": "on_false1", "target": "on_false2", "map_all_data": True},
        {"source": "on_true1", "target": "on_true2", "map_all_data": True},
        {"source": "on_false2", "target": "merge", "map_all_data": True},
        {"source": "on_true2", "target": "merge", "map_all_data": True},
        {"source": "merge", "target": "end", "map_all_data": True},
    ]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})

    assert link_is_required(taskgraph.graph, "start", "fan")
    assert not link_is_required(taskgraph.graph, "fan", "on_false1")
    assert not link_is_required(taskgraph.graph, "fan", "on_true1")
    assert not link_is_required(taskgraph.graph, "on_false1", "on_false2")
    assert not link_is_required(taskgraph.graph, "on_true1", "on_true2")
    assert not link_is_required(taskgraph.graph, "on_false2", "merge")
    assert not link_is_required(taskgraph.graph, "on_true2", "merge")
    assert not link_is_required(
        taskgraph.graph, "merge", "end"
    )  # TODO: this should be True because branches merge again


def test_graph_link_is_required_conditions2():
    nodes = [
        {"id": "start", "task_type": "method", "task_identifier": "dummy"},
        {"id": "fan", "task_type": "method", "task_identifier": "dummy"},
        {"id": "always1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "always2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "merge", "task_type": "method", "task_identifier": "dummy"},
        {"id": "end_always", "task_type": "method", "task_identifier": "dummy"},
    ]
    links = [
        {"source": "start", "target": "fan", "map_all_data": True},
        {"source": "fan", "target": "always1", "map_all_data": True},
        {
            "source": "fan",
            "target": "on_true1",
            "map_all_data": True,
            "conditions": [{"source_output": "result", "value": True}],
        },
        {"source": "always1", "target": "always2", "map_all_data": True},
        {"source": "on_true1", "target": "on_true2", "map_all_data": True},
        {"source": "always2", "target": "merge", "map_all_data": True},
        {"source": "on_true2", "target": "merge", "map_all_data": True},
        {"source": "merge", "target": "end_always", "map_all_data": True},
    ]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})

    assert link_is_required(taskgraph.graph, "start", "fan")
    assert link_is_required(taskgraph.graph, "fan", "always1")
    assert not link_is_required(taskgraph.graph, "fan", "on_true1")
    assert link_is_required(taskgraph.graph, "always1", "always2")
    assert not link_is_required(taskgraph.graph, "on_true1", "on_true2")
    assert link_is_required(taskgraph.graph, "always2", "merge")
    assert not link_is_required(taskgraph.graph, "on_true2", "merge")
    assert not link_is_required(
        taskgraph.graph, "merge", "end_always"
    )  # TODO: this should be True because branches merge again


def test_graph_link_is_required_errors():
    nodes = [
        {"id": "start", "task_type": "method", "task_identifier": "dummy"},
        {"id": "fan", "task_type": "method", "task_identifier": "dummy"},
        {"id": "always1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_error1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "always2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_error2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "merge", "task_type": "method", "task_identifier": "dummy"},
        {"id": "end_always", "task_type": "method", "task_identifier": "dummy"},
        {"id": "end_on_error", "task_type": "method", "task_identifier": "dummy"},
    ]
    links = [
        {"source": "start", "target": "fan", "map_all_data": True},
        {"source": "fan", "target": "always1", "map_all_data": True},
        {
            "source": "fan",
            "target": "on_true1",
            "map_all_data": True,
            "conditions": [{"source_output": "result", "value": True}],
        },
        {
            "source": "fan",
            "target": "on_error1",
            "map_all_data": True,
            "on_error": True,
        },
        {"source": "always1", "target": "always2", "map_all_data": True},
        {"source": "on_true1", "target": "on_true2", "map_all_data": True},
        {"source": "on_error1", "target": "on_error2", "map_all_data": True},
        {"source": "always2", "target": "merge", "map_all_data": True},
        {"source": "on_true2", "target": "merge", "map_all_data": True},
        {"source": "on_error2", "target": "merge", "map_all_data": True},
        {"source": "merge", "target": "end_always", "map_all_data": True},
        {"source": "on_error2", "target": "end_on_error", "map_all_data": True},
    ]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})

    assert link_is_required(taskgraph.graph, "start", "fan")
    assert link_is_required(taskgraph.graph, "fan", "always1")
    assert not link_is_required(taskgraph.graph, "fan", "on_true1")
    assert not link_is_required(taskgraph.graph, "fan", "on_error1")
    assert not link_is_required(taskgraph.graph, "always1", "always2")
    assert not link_is_required(taskgraph.graph, "on_true1", "on_true2")
    assert not link_is_required(taskgraph.graph, "on_error1", "on_error2")
    assert not link_is_required(taskgraph.graph, "always2", "merge")
    assert not link_is_required(taskgraph.graph, "on_true2", "merge")
    assert not link_is_required(taskgraph.graph, "on_error2", "merge")
    assert not link_is_required(
        taskgraph.graph, "merge", "end_always"
    )  # TODO: this should be True because branches merge again
    assert not link_is_required(taskgraph.graph, "on_error2", "end_on_error")



from ewokscore.graph import load_graph
from ewokscore.graph.analysis import start_nodes
from ewokscore.graph.analysis import end_nodes


def test_graph_start_end_nodes():
    nodes = [
        {"id": "start1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "start2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "begin", "task_type": "method", "task_identifier": "dummy"},
        {"id": "fan", "task_type": "method", "task_identifier": "dummy"},
        {"id": "always1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_error1", "task_type": "method", "task_identifier": "dummy"},
        {"id": "always2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_true2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "on_error2", "task_type": "method", "task_identifier": "dummy"},
        {"id": "merge", "task_type": "method", "task_identifier": "dummy"},
        {"id": "end_always", "task_type": "method", "task_identifier": "dummy"},
        {"id": "end_on_error", "task_type": "method", "task_identifier": "dummy"},
    ]
    links = [
        {"source": "start1", "target": "begin", "map_all_data": True},
        {"source": "start2", "target": "begin", "map_all_data": True},
        {
            "source": "begin",
            "target": "start2",
            "map_all_data": True,
            "conditions": [{"source_output": "result", "value": True}],
        },
        {"source": "begin", "target": "fan", "map_all_data": True},
        {"source": "fan", "target": "always1", "map_all_data": True},
        {
            "source": "fan",
            "target": "on_true1",
            "map_all_data": True,
            "conditions": [{"source_output": "result", "value": True}],
        },
        {
            "source": "fan",
            "target": "on_error1",
            "map_all_data": True,
            "on_error": True,
        },
        {"source": "always1", "target": "always2", "map_all_data": True},
        {"source": "on_true1", "target": "on_true2", "map_all_data": True},
        {"source": "on_error1", "target": "on_error2", "map_all_data": True},
        {"source": "always2", "target": "merge", "map_all_data": True},
        {"source": "on_true2", "target": "merge", "map_all_data": True},
        {"source": "on_error2", "target": "merge", "map_all_data": True},
        {"source": "merge", "target": "end_always", "map_all_data": True},
        {"source": "on_error2", "target": "end_on_error", "map_all_data": True},
    ]
    taskgraph = load_graph({"graph": {"id": "test"}, "nodes": nodes, "links": links})

    assert start_nodes(taskgraph.graph) == {"start1"}
    assert end_nodes(taskgraph.graph) == {"end_always", "end_on_error"}



from .sumtask import SumTask


class CondSumTask(SumTask, output_names=["too_small"]):
    """Check whether a value is too small"""

    def run(self):
        super().run()
        self.outputs.too_small = self.outputs.result < 10



import pytest
from ewokscore import load_graph
from ewokscore.graph import graph_io


@pytest.fixture(scope="module")
def graph():
    nodes = [
        {
            "id": "task1",
            "label": "a",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": "mypackage.mymodule.Task1",
        },
        {
            "id": "task2",
            "label": "a",
            "default_inputs": [{"name": "a", "value": 2}],
            "task_type": "class",
            "task_identifier": "mypackage.mymodule.Task2",
        },
        {
            "id": "task3",
            "label": "b",
            "default_inputs": [{"name": "a", "value": 3}],
            "task_type": "class",
            "task_identifier": "mypackage.mymodule.Task1",
        },
        {
            "id": "task4",
            "label": "c",
            "default_inputs": [{"name": "a", "value": 4}],
            "task_type": "class",
            "task_identifier": "mypackage.mymodule.Task2",
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task2",
        },
        {
            "source": "task2",
            "target": "task3",
        },
        {
            "source": "task3",
            "target": "task4",
        },
    ]

    graph = {
        "graph": {"id": "testgraph"},
        "links": links,
        "nodes": nodes,
    }
    return load_graph(graph).graph


def test_parse_inputs(graph):
    inputs = list()
    expected = inputs
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"id": "task1", "name": "a", "value": 10}]
    expected = inputs
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"label": "b", "name": "a", "value": 10}]
    expected = [{"id": "task3", "name": "a", "value": 10}]
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"label": "a", "name": "a", "value": 10}]
    expected = [
        {"id": "task1", "name": "a", "value": 10},
        {"id": "task2", "name": "a", "value": 10},
    ]
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"task_identifier": "Task1", "name": "a", "value": 10}]
    expected = [
        {"id": "task1", "name": "a", "value": 10},
        {"id": "task3", "name": "a", "value": 10},
    ]
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"task_identifier": "Task1", "label": "a", "name": "a", "value": 10}]
    expected = [{"id": "task1", "name": "a", "value": 10}]
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"name": "a", "value": 10}]
    expected = [{"id": "task1", "name": "a", "value": 10}]
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected

    inputs = [{"all": True, "name": "a", "value": 10}]
    expected = [
        {"id": "task1", "name": "a", "value": 10},
        {"id": "task2", "name": "a", "value": 10},
        {"id": "task3", "name": "a", "value": 10},
        {"id": "task4", "name": "a", "value": 10},
    ]
    inputs = graph_io.parse_inputs(graph, inputs)
    assert inputs == expected


def test_update_default_inputs(graph):
    expected = {
        "task1": [{"name": "a", "value": 1}],
        "task2": [{"name": "a", "value": 2}],
        "task3": [{"name": "a", "value": 3}],
        "task4": [{"name": "a", "value": 4}],
    }
    graph_io.update_default_inputs(graph, list())
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"id": "task1", "name": "a", "value": 10}]
    expected["task1"] = [{"name": "a", "value": 10}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"label": "b", "name": "a", "value": 20}]
    expected["task3"] = [{"name": "a", "value": 20}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"label": "a", "name": "a", "value": 30}]
    expected["task1"] = [{"name": "a", "value": 30}]
    expected["task2"] = [{"name": "a", "value": 30}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"task_identifier": "Task1", "name": "a", "value": 40}]
    expected["task1"] = [{"name": "a", "value": 40}]
    expected["task3"] = [{"name": "a", "value": 40}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"task_identifier": "Task1", "label": "a", "name": "a", "value": 50}]
    expected["task1"] = [{"name": "a", "value": 50}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"name": "a", "value": 60}]
    expected["task1"] = [{"name": "a", "value": 60}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]

    inputs = [{"all": True, "name": "a", "value": 70}]
    expected["task1"] = [{"name": "a", "value": 70}]
    expected["task2"] = [{"name": "a", "value": 70}]
    expected["task3"] = [{"name": "a", "value": 70}]
    expected["task4"] = [{"name": "a", "value": 70}]
    graph_io.update_default_inputs(graph, inputs)
    for node_id, node_attrs in graph.nodes.items():
        assert node_attrs["default_inputs"] == expected[node_id]


def test_parse_outputs(graph):
    outputs = None
    expected = [{"id": "task4"}]
    outputs = graph_io.parse_outputs(graph, outputs)
    assert outputs == expected

    outputs = [{"all": False}]
    expected = [{"id": "task4"}]
    outputs = graph_io.parse_outputs(graph, outputs)
    assert outputs == expected

    outputs = [{"all": True}]
    expected = [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}, {"id": "task4"}]
    outputs = graph_io.parse_outputs(graph, outputs)
    assert outputs == expected

    outputs = [{"all": False}]
    expected = [{"id": "task4"}]
    outputs = graph_io.parse_outputs(graph, outputs)
    assert outputs == expected



import networkx

from ..utils import dict_merge


def flatten_multigraph(graph: networkx.DiGraph) -> networkx.DiGraph:
    """The attributes of links between the same two nodes are merged."""
    if not graph.is_multigraph():
        return graph
    newgraph = networkx.DiGraph(**graph.graph)

    edgeattrs = dict()
    for edge, attrs in graph.edges.items():
        key = edge[:2]
        mergedattrs = edgeattrs.setdefault(key, dict())
        # mergedattrs["links"] and attrs["links"]
        # could be two sequences that need to be concatenated
        dict_merge(mergedattrs, attrs, contatenate_sequences=True)

    for name, attrs in graph.nodes.items():
        newgraph.add_node(name, **attrs)
    for (source_id, target_id), mergedattrs in edgeattrs.items():
        newgraph.add_edge(source_id, target_id, **mergedattrs)
    return newgraph



from typing import Callable, Optional
import networkx
from packaging.version import parse as parse_version
import logging
import importlib

# Major version: increment when changing the existing schema
# Minor version: increment when adding features or deprecating the existing schema
DEFAULT_VERSION = parse_version("1.0")
LATEST_VERSION = parse_version("1.0")

# Map graph versions to ewokscore version bounds. Whenever we change the schema
# which the current ewokscore version needs and updating is not possible:
#   - increment the ewokscore version
#   - use that version as upper bound of the last item of _VERSION_BOUNDS
#   - use that version as lower bound of a new item of _VERSION_BOUNDS
_VERSION_BOUNDS = None


def get_version_bounds() -> dict:
    global _VERSION_BOUNDS
    if _VERSION_BOUNDS:
        return _VERSION_BOUNDS

    _VERSION_BOUNDS = dict()
    _VERSION_BOUNDS[parse_version("0.0")] = parse_version("0.0"), parse_version("0.0.1")
    _VERSION_BOUNDS[parse_version("0.1")] = parse_version("0.1.0-rc"), None
    _VERSION_BOUNDS[parse_version("0.2")] = parse_version("0.1.0-rc"), None
    _VERSION_BOUNDS[parse_version("1.0")] = parse_version("0.1.0-rc"), None
    return _VERSION_BOUNDS


logger = logging.getLogger(__name__)


def normalize_schema_version(graph: dict):
    schema_version = graph["graph"].get("schema_version", None)
    if not schema_version:
        schema_version = DEFAULT_VERSION
        graph["graph"]["schema_version"] = str(schema_version)
        logger.info(
            'Graph has no "schema_version": assume version "%s"', schema_version
        )
        return
    pversion = parse_version(schema_version)
    if pversion != LATEST_VERSION:
        # This warning is given because an exception may occur before `update_graph_schema`
        # is called due to the different schema version.
        logger.warning(
            'Graph schema version "%s" is not equal to the latest version "%s"',
            pversion,
            LATEST_VERSION,
        )
    graph["graph"]["schema_version"] = str(pversion)


def update_graph_schema(graph: networkx.DiGraph) -> bool:
    """Updates the graph description to a higher schema version (returns `True`) or raises an
    exception. If the schema version is known it will provide library version bounds
    in the exception message. Returns `False` when the graph does not need
    any update.
    """
    schema_version = graph.graph.get("schema_version", None)
    if schema_version is None:
        schema_version = DEFAULT_VERSION
        graph.graph["schema_version"] = str(schema_version)
        logger.info(
            'Graph has no "schema_version": assume version "%s"', schema_version
        )
    else:
        schema_version = parse_version(schema_version)
    if schema_version == LATEST_VERSION:
        return False

    update_method = _get_update_method(schema_version)
    if update_method:
        before = graph.graph.get("schema_version", None)
        try:
            update_method(graph)
        except Exception:
            pass  # version is not longer supported
        else:
            after = graph.graph.get("schema_version", None)
            assert before != after, "graph conversion did not update the schema version"
            return True

    lbound, ubound = get_version_bounds().get(schema_version, (None, None))
    if lbound and ubound:
        raise ValueError(
            f'Graph schema version "{schema_version}" requires another library version: python3 -m pip install "ewokscore>={lbound},<{ubound}"`'
        )
    elif lbound:
        raise ValueError(
            f'Graph schema version "{schema_version}" requires another library version: python3 -m pip install "ewokscore>={lbound}"'
        )
    elif ubound:
        raise ValueError(
            f'Graph schema version "{schema_version}" requires another library version: python3 -m pip install "ewokscore<{ubound}"'
        )
    else:
        raise ValueError(
            f'Graph schema version "{schema_version}" is either invalid or requires a newer library version: python3 -m pip install --upgrade ewokscore'
        )


def _get_update_method(schema_version) -> Optional[Callable[[networkx.DiGraph], None]]:
    try:
        mod = importlib.import_module(
            __name__ + ".v" + str(schema_version).replace(".", "_")
        )
    except ImportError:
        return None
    return mod.update_graph_schema



import gc
import numpy
from ewokscore import hashing


def test_hashing_unique():
    unique_values = [
        None,
        "",
        b"",
        "abc",
        b"abc",
        0,
        1,
        0.0,
        1.0,
        numpy.int64(0),
        numpy.int32(1),
    ]
    for v in unique_values:
        assert hashing.uhash(v) == hashing.uhash(v)
    assert len({hashing.uhash(v) for v in unique_values}) == len(unique_values)

    alist = list(unique_values)
    assert hashing.uhash(alist) == hashing.uhash(list(alist))
    assert hashing.uhash(alist) != hashing.uhash(alist[::-1])
    assert hashing.uhash(alist) != hashing.uhash(tuple(alist))
    assert hashing.uhash(alist) != hashing.uhash(set(alist))

    aset = set(unique_values)
    assert hashing.uhash(aset) == hashing.uhash(set(aset))
    assert hashing.uhash(aset) != hashing.uhash(tuple(aset))
    assert hashing.uhash(aset) != hashing.uhash(list(alist))

    andarray = numpy.arange(10)
    assert hashing.uhash(andarray) == hashing.uhash(andarray.copy())
    assert hashing.uhash(andarray) != hashing.uhash(andarray.tolist())

    adict = {-i: v for i, v in enumerate(unique_values, 1)}
    assert hashing.uhash(adict) == hashing.uhash(adict)
    assert hashing.uhash(adict) == hashing.uhash(dict(sorted(adict.items())))


def test_pre_uhash():
    class Test(hashing.UniversalHashable, version=1):
        def __init__(self, data, **kw):
            self.data = data
            super().__init__(**kw)

        def _uhash_data(self):
            return self.data

    a = Test(1, pre_uhash=None, instance_nonce=10)
    b = Test(2, pre_uhash=a)  # keeps a reference
    c = Test(3, pre_uhash=a.uhash)  # no reference

    assert a.uhash == b.uhash
    assert a.uhash == c.uhash

    a.data = 3

    assert a.uhash == b.uhash
    assert a.uhash != c.uhash


def test_versioning():
    def create_classes(vparent=None, vchild=None):
        class Mixin1:
            pass

        class Mixin2:
            pass

        class Parent(Mixin1, hashing.UniversalHashable, Mixin2, version=vparent):
            pass

        class Child(Parent, version=vchild):
            pass

        return Parent.class_nonce(), Child.class_nonce()

    vparent1, vchild1 = create_classes()
    vparent2, vchild2 = create_classes(vparent=1)
    vparent3, vchild3 = create_classes(vchild=1)
    vparent4, vchild4 = create_classes(vparent=1, vchild=1)

    assert vparent1 != vchild1
    assert vparent2 != vchild2
    assert vparent3 != vchild3
    assert vparent4 != vchild4

    assert vparent1 != vparent2
    assert vparent1 == vparent3
    assert vparent2 == vparent4

    assert vchild1 != vchild2
    assert vchild2 != vchild3
    assert vchild3 != vchild4


def test_uhash_fixing():
    class Test(hashing.UniversalHashable):
        def __init__(self, data):
            self.data = data
            super().__init__()

        def _uhash_data(self):
            return self.data

    data = [0, 1, 2]
    var = Test(data)

    uhash1 = uhash1org = var.uhash
    data[0] += 1
    uhash2 = var.uhash
    assert uhash1 != uhash2

    uhash1 = var.uhash
    var.fix_uhash()
    uhash2 = var.uhash
    data[0] += 1
    uhash3 = var.uhash
    assert uhash1 == uhash2
    assert uhash1 == uhash3

    uhash1 = var.uhash
    var.undo_fix_uhash()
    uhash2 = var.uhash
    assert uhash1 != uhash2

    data[0] = 0
    uhash = var.uhash
    assert uhash1org == uhash


def test_hashable_cleanup_references():
    class Myclass(hashing.UniversalHashable):
        def __init__(self, data, **kw):
            self.data = data
            super().__init__(**kw)

        def _uhash_data(self):
            return self.data

    obj1 = Myclass(10)
    nref_start = len(gc.get_referrers(obj1))
    obj2 = Myclass(10, pre_uhash=obj1)
    assert len(gc.get_referrers(obj1)) > nref_start

    obj1.data += 1
    assert obj1.uhash == obj2.uhash

    obj2.cleanup_references()
    while gc.collect():
        pass
    assert len(gc.get_referrers(obj1)) == nref_start
    assert obj1.uhash == obj2.uhash

    obj1.data += 1
    assert obj1.uhash != obj2.uhash



import argparse
from ewokscore import cliutils


def test_cli_execute_workflow():
    parser = argparse.ArgumentParser()
    cliutils.add_execute_parameters(parser)
    argv = [
        "acyclic1",
        "--test",
        "-p",
        "a=1",
        "-p",
        "task1:b=test",
        "--workflow-dir",
        "/tmp",
    ]
    args = parser.parse_args(argv)
    cliutils.apply_execute_parameters(args)

    assert args.graph["graph"]["id"] == "acyclic1"

    execute_options = {
        "inputs": [
            {"all": False, "name": "a", "value": 1},
            {"id": "task1", "name": "b", "value": "test"},
        ],
        "merge_outputs": False,
        "outputs": [],
        "varinfo": {"root_uri": "", "scheme": "nexus"},
        "load_options": {"root_dir": "/tmp"},
        "execinfo": {},
    }
    assert args.execute_options == execute_options


def test_cli_convert_workflow():
    parser = argparse.ArgumentParser()
    cliutils.add_convert_parameters(parser)
    argv = [
        "acyclic1",
        "test.json",
        "--test",
        "-p",
        "a=1",
        "-p",
        "task1:b=test",
        "--src-format",
        "yaml",
        "--dst-format",
        "json",
    ]
    args = parser.parse_args(argv)
    cliutils.apply_convert_parameters(args)

    assert args.graph["graph"]["id"] == "acyclic1"

    convert_options = {
        "inputs": [
            {"all": False, "name": "a", "value": 1},
            {"id": "task1", "name": "b", "value": "test"},
        ],
        "load_options": {"representation": "yaml"},
        "save_options": {"representation": "json"},
    }
    assert args.convert_options == convert_options



"""Manage the ewoks event logger which is a global object
"""

import os
import logging
from contextlib import contextmanager
from typing import Dict, Iterable, List, Optional, Tuple

from ewoksutils.logging_utils.asyncwrapper import AsyncHandlerWrapper
from ewoksutils.logging_utils.cleanup import cleanup_logger

from .handlers import is_ewoks_event_handler, instantiate_handler

EWOKS_EVENT_LOGGER_NAME = __name__
ENABLE_EWOKS_EVENTS_BY_DEFAULT = True


def send(
    *args,
    handlers: Optional[List[Dict[str, str]]] = None,
    asynchronous: Optional[bool] = None,
    **kw
) -> None:
    """Send an event to the global ewoks event logger."""
    with _ewoks_event_logger(handlers=handlers, asynchronous=asynchronous) as logger:
        logger.info(*args, **kw)


def add_handler(
    handler,
    asynchronous: Optional[bool] = None,
) -> None:
    """Add a handler to the global ewoks event logger."""
    with _ewoks_event_logger() as logger:
        if _has_handler_instance(logger, handler):
            return
        if asynchronous is None and is_ewoks_event_handler(handler):
            asynchronous = handler.BLOCKING
        if asynchronous:
            handler = AsyncHandlerWrapper(handler)
        logger.addHandler(handler)


def remove_handler(handler: logging.Handler) -> None:
    """Remove a handler from all loggers that receive ewoks event."""
    with _ewoks_event_logger() as logger:
        for linstance, hinstance in _iter_handler_owners(logger, handler):
            linstance.removeHandler(hinstance)
            hinstance.close()


def cleanup():
    """Pending events will be dropped"""
    logging._acquireLock()
    try:
        _cleanup_ewoks_event_logger()
    finally:
        logging._releaseLock()


def _after_fork_in_child():
    cleanup()


if hasattr(os, "register_at_fork"):
    os.register_at_fork(after_in_child=_after_fork_in_child)


@contextmanager
def _ewoks_event_logger(
    handlers: Optional[List[Dict[str, str]]] = None, asynchronous: Optional[bool] = None
) -> Iterable[logging.Logger]:
    """Initialize and yield the ewoks event logger"""
    # Issue with logging and forking:
    # https://pythonspeed.com/articles/python-multiprocessing/

    logging._acquireLock()
    try:
        if _ewoks_event_logger_requires_cleanup():
            _cleanup_ewoks_event_logger()
        if _ewoks_event_logger_requires_init():
            _init_ewoks_event_logger(handlers, asynchronous)
        yield logging.getLogger(EWOKS_EVENT_LOGGER_NAME)
    finally:
        logging._releaseLock()


def _cleanup_ewoks_event_logger():
    """Cleanup and delete the global ewoks event logger"""
    cleanup_logger(EWOKS_EVENT_LOGGER_NAME)


def _ewoks_event_logger_requires_init() -> bool:
    logger = logging.getLogger(EWOKS_EVENT_LOGGER_NAME)
    return not hasattr(logger, "ewoks_pid")


def _ewoks_event_logger_requires_cleanup() -> bool:
    logger = logging.getLogger(EWOKS_EVENT_LOGGER_NAME)
    ewoks_pid = getattr(logger, "ewoks_pid", None)
    return ewoks_pid is not None and ewoks_pid != os.getpid()


def _init_ewoks_event_logger(
    handlers: Optional[List[Dict[str, str]]], asynchronous: Optional[bool]
):
    logger = logging.getLogger(EWOKS_EVENT_LOGGER_NAME)
    logger.setLevel(logging.DEBUG)
    logger.ewoks_pid = os.getpid()
    if not handlers:
        return
    for desc in handlers:
        try:
            kwargs = {
                arg["name"]: arg["value"] for arg in desc.get("arguments", list())
            }
            handler = instantiate_handler(desc["class"], **kwargs)
        except Exception as e:
            raise RuntimeError(
                "cannot create an ewoks event handler from the description"
            ) from e
        asynchronous_handler = desc.get("asynchronous", asynchronous)
        add_handler(handler, asynchronous=asynchronous_handler)


def _iter_loggers(logger: logging.Logger) -> Iterable[logging.Logger]:
    """Yield all loggers which will receive ewoks events."""
    _logger = logger
    while _logger is not None:
        yield _logger
        if not _logger.propagate:
            return
        _logger = _logger.parent


def _iter_handler_owners(
    logger: logging.Logger, instance: logging.Handler
) -> Iterable[Tuple[logging.Logger, logging.Handler]]:
    """Yield all loggers which have a specific handler (or a handler that wraps the specific event handler)."""
    for _logger in _iter_loggers(logger):
        for handler in _logger.handlers:
            if handler is instance:
                yield _logger, instance
            elif isinstance(handler, AsyncHandlerWrapper):
                instance2 = handler.wrapped_handler
                if instance2 is instance:
                    yield _logger, instance2


def _has_handler_instance(logger: logging.Logger, instance: logging.Handler) -> bool:
    """Is this handler registered with a logger or it's parents?"""
    for _ in _iter_handler_owners(logger, instance):
        return True
    return False



"""Contexts for job, workflow or node events. Allows for initializing
event fields and sending start/end events.
"""

from contextlib import contextmanager
from functools import wraps
from typing import Mapping, Union
from .send_events import send_job_event
from .send_events import send_workflow_event
from .send_events import ExecInfoType
from .initialize_events import init_job
from .initialize_events import init_workflow
from .initialize_events import init_node
from . import global_state


RawExecInfoType = Union[Mapping, bool, str, None]


def job_decorator(**static_job_info):
    def _job_decorator(method):
        @wraps(method)
        def wrapper(*args, execinfo: RawExecInfoType = None, **kw):
            with job_context(execinfo, **static_job_info) as execinfo:
                return method(*args, execinfo=execinfo, **kw)

        return wrapper

    return _job_decorator


@contextmanager
def job_context(execinfo: RawExecInfoType, **static_job_info) -> ExecInfoType:
    if execinfo is None:
        execinfo = global_state.ENABLE_EWOKS_EVENTS_BY_DEFAULT

    if isinstance(execinfo, str):
        execinfo = {"job_id": execinfo}
    elif isinstance(execinfo, bool):
        if execinfo:
            execinfo = dict()
        else:
            execinfo = None
    elif execinfo is None:
        pass
    elif not isinstance(execinfo, Mapping):
        raise TypeError

    execinfo = init_job(execinfo, **static_job_info)
    if execinfo is None:
        yield None
    else:
        with _context(execinfo, "job", send_job_event, execinfo["job_id"]) as execinfo:
            yield execinfo


@contextmanager
def workflow_context(execinfo: ExecInfoType, **kw) -> ExecInfoType:
    execinfo = init_workflow(execinfo, **kw)
    if execinfo is None:
        yield None
    else:
        with _context(
            execinfo, "workflow", send_workflow_event, execinfo["workflow_id"]
        ) as execinfo:
            yield execinfo


@contextmanager
def node_context(execinfo: ExecInfoType, **kw) -> ExecInfoType:
    yield init_node(execinfo, **kw)


@contextmanager
def _context(
    execinfo: ExecInfoType, context, send_context_event, obj_id
) -> ExecInfoType:
    contexts = execinfo.get("contexts")
    if contexts is None:
        contexts = {"job": list(), "workflow": list()}
        execinfo["contexts"] = contexts
    obj_ids = contexts[context]
    first_context = obj_id not in obj_ids
    if first_context:
        send_context_event(execinfo=execinfo, event="start")
        obj_ids.append(obj_id)
    try:
        yield execinfo
    except BaseException as e:
        if not execinfo.get("exception"):
            execinfo["exception"] = e
        raise
    finally:
        if first_context:
            obj_ids.remove(obj_id)
            send_context_event(execinfo=execinfo, event="end")



from .task import Task
from ewoksutils.import_utils import import_method

METHOD_ARGUMENT = "_method"


class MethodExecutorTask(
    Task, input_names=[METHOD_ARGUMENT], output_names=["return_value"]
):
    METHOD_ARGUMENT = METHOD_ARGUMENT

    def run(self):
        kwargs = self.get_named_input_values()
        args = self.get_positional_input_values()
        fullname = kwargs.pop(self.METHOD_ARGUMENT)
        method = import_method(fullname)

        result = method(*args, **kwargs)

        self.outputs.return_value = result



"""Define and parse ewoks events
"""

from numbers import Number
from typing import Dict, List, Mapping, Optional
from datetime import datetime
import traceback
from warnings import warn
from ewoksutils.event_utils import FIELD_TYPES

from . import global_state
from .initialize_events import ExecInfoType


def send_job_event(**kw):
    kw = _preprocess_event(**kw)
    logargs, logkwargs = _parse_job_event(**kw)
    _send_event(*logargs, **logkwargs)


def send_workflow_event(**kw):
    kw = _preprocess_event(**kw)
    logargs, logkwargs = _parse_workflow_event(**kw)
    _send_event(*logargs, **logkwargs)


def send_task_event(**kw):
    kw = _preprocess_event(**kw)
    logargs, logkwargs = _parse_task_event(**kw)
    _send_event(*logargs, **logkwargs)


def timestamp() -> str:
    return datetime.now().astimezone().isoformat()


def _preprocess_event(execinfo: ExecInfoType = None, **logkwargs):
    if execinfo:
        logkwargs = {**execinfo, **logkwargs}
    if not logkwargs.get("time"):
        logkwargs["time"] = timestamp()
    _parse_exception(logkwargs)
    return logkwargs


def _parse_exception(logkwargs):
    exception = logkwargs.pop("exception", None)
    if exception is None:
        return
    elif isinstance(exception, str):
        logkwargs["error"] = True
        tb = exception
        error_message = _extract_reason_from_traceback(tb)
        if not logkwargs.get("error_message"):
            logkwargs["error_message"] = error_message
        if not logkwargs.get("error_traceback"):
            logkwargs["error_traceback"] = tb
    elif isinstance(exception, BaseException):
        logkwargs["error"] = True
        tb = "".join(
            traceback.format_exception(
                type(exception), exception, exception.__traceback__
            )
        )
        if not logkwargs.get("error_traceback"):
            logkwargs["error_traceback"] = tb
        if logkwargs.get("context") == "node":
            exception = _first_exception(exception)
        if not logkwargs.get("error_message"):
            error_message = str(exception)
            if not error_message:
                error_message = type(exception).__name__
            logkwargs["error_message"] = error_message
    else:
        raise TypeError(
            "ewoks event field 'exception' should be a string or an exception instance"
        )


def _first_exception(exception: BaseException) -> BaseException:
    while True:
        if exception.__cause__ is not None:
            # exception raised from another exception
            exception = exception.__cause__
        elif exception.__context__ is not None:
            # exception raised during handling of another exception
            exception = exception.__context__
        else:
            return exception


def _extract_reason_from_traceback(tb: str) -> str:
    for line in reversed(tb.split("\n")):
        if ":" in line:
            return line
    return ""


def _parse_workflow_event(**logkwargs):
    event_data, logkwargs = _extract_workflow_fields(**logkwargs)
    msg = _validate_event(event_data)
    logargs = (
        "[job %r] [workflow %r] %s",
        event_data["job_id"],
        event_data["workflow_id"],
        msg,
    )
    logkwargs["event_data"] = event_data
    return logargs, logkwargs


def _parse_job_event(**logkwargs):
    event_data, logkwargs = _extract_job_fields(**logkwargs)
    msg = _validate_event(event_data)
    logargs = ("[job %r] %s", event_data["job_id"], msg)
    logkwargs["event_data"] = event_data
    return logargs, logkwargs


def _parse_task_event(**logkwargs):
    event_data, logkwargs = _extract_task_fields(**logkwargs)
    msg = _validate_event(event_data)
    logargs = (
        "[job %r] [workflow %r] [node %r] [task %r] %s",
        event_data["job_id"],
        event_data["workflow_id"],
        event_data["node_id"],
        event_data["task_id"],
        msg,
    )
    logkwargs["event_data"] = event_data
    return logargs, logkwargs


def _extract_common_fields(
    host_name: str,
    process_id: int,
    user_name: str,
    job_id: str,
    event: str,
    engine: Optional[str] = None,
    binding: Optional[str] = None,
    time: Optional[str] = None,
    error: Optional[bool] = None,
    error_message: Optional[str] = None,
    error_traceback: Optional[str] = None,
    **logkwargs,
):
    if "engine" in FIELD_TYPES:
        engine_field = "engine"
    else:
        engine_field = "binding"
    if binding:
        if engine:
            raise ValueError("'binding' and 'engine' cannot be used together")
        engine = binding
        if engine_field == "engine":
            warn("'binding' is deprecated in favor of 'engine'", DeprecationWarning)
    event_data = {
        "host_name": host_name,
        "process_id": process_id,
        "user_name": user_name,
        "job_id": job_id,
        engine_field: engine,
        "type": event,
        "time": time,
        "error": error,
        "error_message": error_message,
        "error_traceback": error_traceback,
    }
    return event_data, logkwargs


def _extract_job_fields(
    **logkwargs,
):
    event_data, logkwargs = _extract_common_fields(**logkwargs)
    event_data["context"] = "job"
    return event_data, logkwargs


def _extract_workflow_fields(
    workflow_id: str,
    **logkwargs,
):
    event_data, logkwargs = _extract_job_fields(**logkwargs)
    event_data["context"] = "workflow"
    event_data["workflow_id"] = workflow_id
    return event_data, logkwargs


def _extract_task_fields(
    node_id: str,
    task_id: str,
    progress: Optional[Number] = None,
    task_uri: Optional[str] = None,
    input_uris: Optional[List[Dict[str, Optional[str]]]] = None,
    output_uris: Optional[List[Dict[str, Optional[str]]]] = None,
    **logkwargs,
):
    event_data, logkwargs = _extract_workflow_fields(**logkwargs)
    event_data["context"] = "node"
    event_data["node_id"] = node_id
    event_data["task_id"] = task_id
    event_data["progress"] = progress
    event_data["task_uri"] = task_uri
    event_data["input_uris"] = input_uris
    event_data["output_uris"] = output_uris
    return event_data, logkwargs


def _validate_event(event_data: dict) -> str:
    event_data.update((field, None) for field in set(FIELD_TYPES) - set(event_data))
    event = event_data["type"]
    if event_data["context"] == "job":
        obj = "job"
    elif event_data["context"] == "workflow":
        obj = "workflow"
    elif event_data["context"] == "node":
        obj = "task"
    else:
        raise ValueError(f"context '{event_data['context']}' is unknown")
    if event == "start":
        return f"{obj} started"
    elif event == "end":
        error = bool(
            event_data["error"]
            or event_data["error_message"]
            or event_data["error_traceback"]
        )
        event_data["error"] = error
        if error:
            if event_data["error_message"] is None:
                event_data["error_message"] = ""
            if event_data["error_traceback"] is None:
                event_data["error_traceback"] = ""
        if not error:
            return f"{obj} finished"
        elif event_data["error"] and event_data["error_message"]:
            return f"{obj} failed ({event_data['error_message']})"
        else:
            return f"{obj} failed"
    elif event == "progress":
        if not isinstance(event_data["progress"], Number):
            event_data["progress"] = float("nan")
        return f"{obj} progress {event_data['progress']}%"
    else:
        raise ValueError(f"unknown ewoks event type '{event}'")


def _send_event(msg: str, *args, event_data=None, extra=None, contexts=None, **kw):
    if not isinstance(event_data, Mapping):
        raise TypeError("'event_data' should be a mapping")
    if extra:
        if not isinstance(extra, Mapping):
            raise TypeError("'extra' should be a mapping")
        extra = {**extra, **event_data}
    else:
        extra = event_data
    global_state.send(msg, *args, extra=extra, **kw)



from typing import Callable, Optional
from .task import Task
from ewoksutils.import_utils import import_method


_GENERATORS = dict()


def get_task_class_generator(generator_qualname: str) -> Callable[[str], Task]:
    if not generator_qualname:
        return Task.get_subclass
    task_class_generator = _GENERATORS.get(generator_qualname, None)
    if task_class_generator is not None:
        return task_class_generator
    task_class_generator = import_method(generator_qualname)
    _GENERATORS[generator_qualname] = task_class_generator
    return task_class_generator


def get_dynamically_task_class(
    generator_qualname: Optional[str], registry_name: str
) -> Task:
    return get_task_class_generator(generator_qualname)(registry_name)



from pprint import pformat as _pformat  # hide from task discovery
from ewokscore import Task


class NoOutputTask(Task):
    """A task without outputs"""

    def run(self):
        input_values = self.get_input_values()
        if input_values:
            print(f"{self.label}: {_pformat(input_values)}")
        else:
            print(f"{self.label}: <no inputs>")



import networkx

from ..inittask import validate_task_executable
from .analysis import required_predecessors
from .schema import update_graph_schema


def validate_graph(graph: networkx.DiGraph) -> None:
    while update_graph_schema(graph):
        pass
    _validate_nodes(graph)
    _validate_links(graph)


def _validate_nodes(graph: networkx.DiGraph) -> None:
    for node_id, node_attrs in graph.nodes.items():
        validate_task_executable(node_id, node_attrs)

        # Isolated nodes do no harm so comment this
        # if len(graph.nodes) > 1 and not node_has_links(graph, node_id):
        #    raise ValueError(f"Node {repr(node_id)} has no links")

        inputs_from_required = dict()
        for source_id in required_predecessors(graph, node_id):
            link_attrs = graph[source_id][node_id]
            arguments = link_attrs.get("data_mapping", list())
            for arg in arguments:
                try:
                    name = arg["target_input"]
                except KeyError:
                    raise KeyError(
                        f"Argument '{arg}' of link '{source_id}' -> '{node_id}' is missing an 'input' key"
                    ) from None
                other_source_id = inputs_from_required.get(name)
                if other_source_id:
                    raise ValueError(
                        f"Node {repr(source_id)} and {repr(other_source_id)} both connect to the input {repr(name)} of {repr(node_id)}"
                    )
                inputs_from_required[name] = source_id


def _validate_links(graph: networkx.DiGraph) -> None:
    for (source, target), linkattrs in graph.edges.items():
        err_msg = f"Link {source}->{target}: '{{}}' and '{{}}' cannot be used together"
        if linkattrs.get("map_all_data") and linkattrs.get("data_mapping"):
            raise ValueError(err_msg.format("map_all_data", "data_mapping"))
        if linkattrs.get("on_error") and linkattrs.get("conditions"):
            raise ValueError(err_msg.format("on_error", "conditions"))



from . import graph


@graph
def acyclic1():
    task = "ewokscore.tests.examples.tasks.sumtask.SumTask"
    nodes = [
        {
            "id": "task1",
            "default_inputs": [{"name": "a", "value": 1}],
            "task_type": "class",
            "task_identifier": task,
            "ows": {"position": "(323.0, 166.0)"},
            "uiProps": {"position": {"x": 323, "y": 166}},
        },
        {
            "id": "task2",
            "default_inputs": [{"name": "a", "value": 2}],
            "task_type": "class",
            "task_identifier": task,
            "ows": {"position": "(317.0, 320.0)"},
            "uiProps": {"position": {"x": 317, "y": 320}},
        },
        {
            "id": "task3",
            "default_inputs": [{"name": "b", "value": 3}],
            "task_type": "class",
            "task_identifier": task,
            "ows": {"position": "(575.0, 164.0)"},
            "uiProps": {"position": {"x": 575, "y": 164}},
        },
        {
            "id": "task4",
            "default_inputs": [{"name": "b", "value": 4}],
            "task_type": "class",
            "task_identifier": task,
            "ows": {"position": "(577.0, 321.0)"},
            "uiProps": {"position": {"x": 577, "y": 321}},
        },
        {
            "id": "task5",
            "default_inputs": [{"name": "b", "value": 5}],
            "task_type": "class",
            "task_identifier": task,
            "ows": {"position": "(808.0, 231.0)"},
            "uiProps": {"position": {"x": 808, "y": 231}},
        },
        {
            "id": "task6",
            "default_inputs": [{"name": "b", "value": 6}],
            "task_type": "class",
            "task_identifier": task,
            "ows": {"position": "(1042.0, 234.0)"},
            "uiProps": {"position": {"x": 1042, "y": 234}},
        },
    ]

    links = [
        {
            "source": "task1",
            "target": "task3",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task2",
            "target": "task4",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task3",
            "target": "task5",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
        {
            "source": "task4",
            "target": "task5",
            "data_mapping": [{"target_input": "b", "source_output": "result"}],
        },
        {
            "source": "task5",
            "target": "task6",
            "data_mapping": [{"source_output": "result", "target_input": "a"}],
        },
    ]

    ows = {
        "annotations": [
            {
                "id": "0",
                "type": "text",
                "params": {
                    "geometry": [92.0, 137.0, 150.0, 67.0],
                    "text": 'Double-click to "trigger" the workflow',
                    "font": {"family": "Ubuntu", "size": 16},
                    "content_type": "text/plain",
                },
            },
            {
                "id": "1",
                "type": "arrow",
                "params": {
                    "geometry": [[222.0, 167.0], [281.0, 168.0]],
                    "color": "#C1272D",
                },
            },
            {
                "id": "2",
                "type": "text",
                "params": {
                    "geometry": [84.0, 299.0, 150.0, 68.0],
                    "text": 'Double-click to\n"trigger" the workflow',
                    "font": {"family": "Ubuntu", "size": 16},
                    "content_type": "text/plain",
                },
            },
            {
                "id": "3",
                "type": "arrow",
                "params": {
                    "geometry": [[217.0, 319.0], [275.0, 319.0]],
                    "color": "#C1272D",
                },
            },
            {
                "id": "4",
                "type": "text",
                "params": {
                    "geometry": [369.0, 10.0, 499.0, 68.0],
                    "text": 'Double-click any node to see the\n - "Execute" button: execute only the task\n - "Trigger" button: execute the task and all downstream tasks',
                    "font": {"family": "Ubuntu", "size": 16},
                    "content_type": "text/plain",
                },
            },
        ]
    }

    graph = {
        "graph": {"ows": ows},
        "links": links,
        "nodes": nodes,
    }

    expected_results = {
        "task1": {"result": 1},
        "task2": {"result": 2},
        "task3": {"result": 4},
        "task4": {"result": 6},
        "task5": {"result": 10},
        "task6": {"result": 16},
    }

    return graph, expected_results



import os
import sys
import subprocess
from .task import Task

SCRIPT_ARGUMENT = "_script"
WIN32 = sys.platform == "win32"


class ScriptExecutorTask(
    Task,
    input_names=[SCRIPT_ARGUMENT],
    optional_input_names=["_capture_output", "_merge_err", "_raise_on_error"],
    output_names=["return_code", "out", "err"],
):
    SCRIPT_ARGUMENT = SCRIPT_ARGUMENT

    def run(self):
        fullname = self.inputs._script
        if not isinstance(fullname, str):
            raise TypeError(fullname, type(fullname))

        # Python or shell script
        is_python = fullname.endswith(".py")

        # Is script executable?
        if os.path.isfile(fullname):
            # existing python or shell script
            fullname = os.path.abspath(fullname)
            if WIN32:
                is_executable = not is_python
            else:
                with open(fullname, "r") as f:
                    is_executable = f.readline().startswith("#!")
        else:
            # command (although it could be a script that does not exist)
            is_executable = True
            fullname = fullname.split(" ")

        # Select executable when fullname itself is not executable
        executable = None
        if not is_executable:
            if is_python:
                executable = sys.executable
            elif not WIN32:
                executable = "bash"

        # Command starts with "[executable] fullname ..."
        cmd = []
        if executable:
            cmd.append(executable)
        if isinstance(fullname, str):
            cmd.append(fullname)
        else:
            cmd.extend(fullname)

        # Script/command arguments
        if is_python:
            # Use full parameter name
            argmarker = "--"
        else:
            # Use getopts-style parameter parsing by the script
            argmarker = "-"
        skip = self.input_names()
        for k, v in self.get_input_values().items():
            if k not in skip:
                cmd.extend((argmarker + k, str(v)))

        # Run
        stdout = stderr = None
        if self.inputs._capture_output:
            stdout = subprocess.PIPE
            if self.inputs._merge_err:
                stderr = subprocess.STDOUT
            else:
                stderr = subprocess.PIPE

        result = subprocess.run(cmd, cwd=os.getcwd(), stdout=stdout, stderr=stderr)
        if self.inputs._raise_on_error:
            result.check_returncode()

        self.outputs.return_code = result.returncode
        if result.stdout:
            self.outputs.out = result.stdout.decode()
        elif self.inputs._capture_output:
            self.outputs.out = ""
        else:
            self.outputs.out = None
        if result.stderr:
            self.outputs.err = result.stderr.decode()
        elif self.inputs._capture_output and not self.inputs._merge_err:
            self.outputs.err = ""
        else:
            self.outputs.err = None



class MissingData:
    def __bool__(self):
        return False

    def __repr__(self):
        return "<MISSING_DATA>"

    def __eq__(self, other) -> bool:
        return isinstance(other, type(self))


MISSING_DATA = MissingData()


def is_missing_data(data):
    """This method solves the following issues when checking whether data is "missing":

    1. `myvar is MISSING_DATA`: problem when `MISSING_DATA` gets copied somehow
    2. `myvar == MISSING_DATA`: problem when `myvar` for example a `numpy` array
    """
    return isinstance(data, MissingData)



from ewokscore import Task


class ErrorSumTask(
    Task, optional_input_names=["a", "b", "raise_error"], output_names=["result"]
):
    """Add two number with intentional exception"""

    def run(self):
        result = self.get_input_value("a", default=0)
        if self.inputs.b:
            result += self.inputs.b
        self.outputs.result = result
        if self.inputs.raise_error:
            raise RuntimeError("Intentional error")



from typing import List, Sequence
import networkx


def graphs_are_equal(graph1: networkx.DiGraph, graph2: networkx.DiGraph) -> bool:
    if not _attrs_are_equal(graph1.graph, graph2.graph):
        return False
    if set(graph1.nodes) != set(graph2.nodes):
        return False
    for node_id, node_attr1 in graph1.nodes.items():
        node_attr2 = graph2.nodes[node_id]
        if not _attrs_are_equal(node_attr1, node_attr2):
            return False
    if set(graph1.edges) != set(graph2.edges):
        return False
    for edge_id, edge_attrs1 in graph1.edges.items():
        edge_attrs2 = graph2.edges[edge_id]
        if not _attrs_are_equal(edge_attrs1, edge_attrs2):
            return False
    return True


def _attrs_are_equal(attrs1: dict, attrs2: dict) -> bool:
    if set(attrs1) != set(attrs2):
        return False
    for name, value1 in attrs1.items():
        value2 = attrs2[name]
        if name in ("input_nodes", "output_nodes"):
            value1 = _items_to_dict(value1, ["id"])
            value2 = _items_to_dict(value2, ["id"])
        elif name == "default_inputs":
            value1 = _items_to_dict(value1, ["name"])
            value2 = _items_to_dict(value2, ["name"])
        elif name == "data_mapping":
            value1 = _items_to_dict(value1, ["source_output", "target_input"])
            value2 = _items_to_dict(value2, ["source_output", "target_input"])
        elif name == "conditions":
            value1 = _items_to_dict(value1, ["source_output"])
            value2 = _items_to_dict(value2, ["source_output"])
        if value1 != value2:
            return False
    return True


def _items_to_dict(items: List[dict], keys: Sequence[str]) -> dict:
    result = dict()
    for item in items:
        result[tuple(item[k] for k in keys)] = item
    return result



import warnings
from typing import Any, Dict, Optional, Union
from numbers import Integral
from collections.abc import Mapping, MutableMapping, Iterable, Sequence

from . import hashing
from . import missing_data
from .persistence import instantiate_data_proxy
from .persistence.proxy import DataProxy
from .persistence.proxy import DataUri


def data_proxy_from_varinfo(
    uhash_source: hashing.UniversalHashable, varinfo
) -> Optional[DataProxy]:
    root_uri = varinfo.get("root_uri")
    if not root_uri and not varinfo.get("has_data_proxy"):
        return
    relative_uri = varinfo.get("relative_uri")
    scheme = varinfo.get("scheme", "json")
    return instantiate_data_proxy(
        scheme=scheme,
        root_uri=root_uri,
        relative_uri=relative_uri,
        uhash_source=uhash_source,
    )


CONTAINER_URI_NAME = "task_results"


class Variable(hashing.UniversalHashable):
    """Has a runtime value (python object) and a persistent value (disk or memory).
    The location of the persistent value is either provided or derived from the
    universal hash of the variable, which itself can be provided or derived from the
    variable data.
    """

    def __init__(
        self,
        value: Any = missing_data.MISSING_DATA,
        metadata: Any = missing_data.MISSING_DATA,
        varinfo: Optional[dict] = None,
        data_uri: Union[DataUri, str, None] = None,
        data_proxy: Optional[DataProxy] = None,
        pre_uhash: Optional[hashing.PreUhashTypes] = None,
        instance_nonce: Optional[Any] = None,
    ):
        if varinfo is None:
            varinfo = dict()
        elif not isinstance(varinfo, Mapping):
            raise TypeError(varinfo, type(varinfo))

        if pre_uhash is None:
            pre_uhash = varinfo.get("uhash_data", None)

        if data_proxy is not None:
            pre_uhash = data_proxy.uri.uhash
            instance_nonce = None
            self._data_proxy = data_proxy
        elif data_uri is not None:
            if isinstance(data_uri, str):
                data_uri = DataUri(data_uri, self)
            else:
                pre_uhash = data_uri.uhash
            instance_nonce = None
            self._data_proxy = instantiate_data_proxy(uri=data_uri)
            if self._data_proxy is None:
                raise ValueError("Invalid URI", data_uri)
        else:
            self._data_proxy = data_proxy_from_varinfo(
                uhash_source=self, varinfo=varinfo
            )

        self._hashing_enabled = bool(varinfo.get("enable_hashing", False))
        self._hashing_enabled |= self._data_proxy is not None

        self._runtime_value = self.MISSING_DATA
        if missing_data.is_missing_data(metadata):
            metadata = dict()
        self._runtime_metadata = metadata

        super().__init__(pre_uhash=pre_uhash, instance_nonce=instance_nonce)
        self.value = value

    def copy_without_references(self):
        """Copy that does not contain references to uhashable objects"""
        kwargs = self.get_uhash_init(serialize=True)
        kwargs["data_proxy"] = self.data_proxy
        return type(self)(value=self.value, **kwargs)

    @property
    def data_proxy(self) -> Optional[DataProxy]:
        return self._data_proxy

    @property
    def data_uri(self) -> Optional[DataUri]:
        if self._data_proxy is None:
            return None
        return self.data_proxy.uri

    @property
    def hashing_enabled(self):
        return self._hashing_enabled

    def _uhash_data(self):
        """The runtime value is used."""
        if self._hashing_enabled:
            return self._runtime_value
        else:
            return super()._uhash_data()

    def __eq__(self, other):
        if isinstance(other, hashing.HasUhash):
            return super().__eq__(other)
        else:
            return self.value == other

    @property
    def value(self):
        if missing_data.is_missing_data(self._runtime_value):
            self.load(raise_error=False)
        return self._runtime_value

    @value.setter
    def value(self, v):
        self._runtime_value = v

    def is_missing(self) -> bool:
        return missing_data.is_missing_data(self.value)

    @property
    def metadata(self) -> dict:
        return self._runtime_metadata

    def dump(self) -> bool:
        """From runtime to persistent value (never overwrite).
        Creating the persistent value needs to be atomic.

        This silently returns when:
        - data persistence is disabled
        - already persisted
        - data does not have a runtime value (MISSING_DATA)
        - non value URI can be constructed
        """
        if (
            self.data_proxy is not None
            and not self.has_persistent_value
            and self.has_runtime_value
        ):
            return self.data_proxy.dump(self.serialize())
        return False

    def load(self, raise_error=True):
        """From persistent to runtime value. This is called when
        try to get the value (lazy loading).

        This silently returns when:
        - data persistence is disabled
        - uri is None (i.e. uhash is None)
        - raise_error=False
        """
        if self.data_proxy is not None:
            data = self.data_proxy.load(raise_error=raise_error)
            if not missing_data.is_missing_data(data):
                self.deserialize(data)

    def serialize(self) -> dict:
        """Serialize data before persistent storage"""
        data = dict(self.metadata)
        data["value"] = self.value
        return data

    def deserialize(self, data: dict):
        """Deserialize data after loading from persistent storage"""
        # When the value is `None`, the backup may not store it (e.g. HDF5)
        self._runtime_value = data.pop("value", None)
        self._runtime_metadata = data

    @property
    def has_persistent_value(self):
        return self._has_persistent_value()

    @property
    def has_runtime_value(self):
        return self._has_runtime_value()

    @property
    def has_value(self):
        return self.has_runtime_value or self.has_persistent_value

    def _has_persistent_value(self):
        return self.data_proxy is not None and self.data_proxy.exists()

    def _has_runtime_value(self):
        return not missing_data.is_missing_data(self._runtime_value)

    def force_non_existing(self):
        while self.has_persistent_value:
            self.uhash_randomize()


class VariableContainer(Variable, Mapping):
    """An immutable mapping of variable identifiers (str or int) to variables (Variable)."""

    def __init__(
        self,
        value: Any = missing_data.MISSING_DATA,
        varinfo: Optional[dict] = None,
        data_uri: Optional[DataUri] = None,
        data_proxy: Optional[DataProxy] = None,
        pre_uhash: Optional[hashing.PreUhashTypes] = None,
        instance_nonce: Optional[Any] = None,
    ):
        varparams = {
            "pre_uhash": pre_uhash,
            "instance_nonce": instance_nonce,
            "varinfo": varinfo,
        }
        self.__add_relative_uri(varparams, path=CONTAINER_URI_NAME)

        self.__varparams = varparams
        self.__npositional_vars = 0
        super().__init__(data_uri=data_uri, data_proxy=data_proxy, **varparams)
        if value:
            self._update(value)

    def fix_uhash(self):
        for var in self.values():
            var.fix_uhash()
        return super().fix_uhash()

    def cleanup_references(self):
        """Remove all references to other hashables.
        Side effect: fixes the uhash when it depends on another hashable.
        """
        for var in self.values():
            var.cleanup_references()
        pre_uhash = self.__varparams.get("pre_uhash")
        if isinstance(pre_uhash, hashing.HasUhash):
            self.__varparams["pre_uhash"] = pre_uhash.uhash
        return super().cleanup_references()

    def copy_without_references(self):
        """The uhash of the copy is fixed thereby remove references to other uhashable objects."""
        return type(self)(
            value={name: var.copy_without_references() for name, var in self.items()},
            **self.__varparams,
        )

    def __getitem__(self, key):
        return self.value[key]

    def _update(self, value):
        if isinstance(value, Mapping):
            value = value.items()
        if not isinstance(value, Iterable):
            raise TypeError(value, type(value))
        for i, tpl in enumerate(value):
            if not isinstance(tpl, Sequence):
                raise TypeError(
                    f"cannot convert dictionary update sequence element #{i} to a sequence"
                )
            if len(tpl) != 2:
                raise ValueError(
                    f"dictionary update sequence element #{i} has length {len(tpl)}; 2 is required"
                )
            self._set_item(*tpl)

    def _set_item(self, key, value):
        key = self._parse_variable_name(key)
        if isinstance(key, int):
            self._fill_missing_positions(key)
        if not self.container_has_value:
            self.value = dict()
        self.value[key] = self._create_variable(key, value)

    def _parse_variable_name(self, key):
        """Variables are identified by a `str` or an `int`. A key like "1" will
        be converted to an `int` (e.g. json dump converts `int` to  `str`).
        """
        if isinstance(key, str):
            if key.isdigit():
                key = int(key)
        if isinstance(key, Integral):
            key = int(key)
            if key < 0:
                raise ValueError("Negative argument positions are not allowed")
        elif not isinstance(key, str):
            raise TypeError(
                f"Variable {key} must be a string or positive integer", type(key)
            )
        return key

    def _fill_missing_positions(self, pos):
        nbefore = self.__npositional_vars
        nafter = max(nbefore, pos + 1)
        for i in range(nbefore, nafter - 1):
            self._set_item(i, self.MISSING_DATA)
        self.__npositional_vars = nafter

    @property
    def n_positional_variables(self):
        return self.__npositional_vars

    def _create_variable(self, name, value):
        if isinstance(value, Variable):
            return value
        varparams = dict()
        if isinstance(value, DataUri):
            varparams["data_uri"] = value
        elif isinstance(value, DataProxy):
            varparams["data_proxy"] = value
        elif isinstance(value, (hashing.UniversalHash, hashing.HasUhash)):
            varparams.update(self.__varparams)
            varparams["pre_uhash"] = value
            varparams["instance_nonce"] = None
        else:
            varparams.update(self.__varparams)
            varparams["value"] = value
            instance_nonce = varparams.pop("instance_nonce", None)
            varparams["instance_nonce"] = instance_nonce, name
            self.__add_relative_uri(varparams, path=f"{name}")
        return Variable(**varparams)

    def __copy_varinfo(self, varparams) -> Dict:
        varinfo = varparams.get("varinfo")
        if varinfo is None:
            varparams["varinfo"] = dict()
        else:
            varparams["varinfo"] = dict(varinfo)
        return varparams["varinfo"]

    def __add_relative_uri(self, varparams, path: Optional[str] = None) -> bool:
        """Modify the general `root_uri` for a specific case."""
        varinfo = varparams.get("varinfo")
        if not varinfo or not varinfo.get("root_uri"):
            return False
        varinfo = self.__copy_varinfo(varparams)
        if path:
            if varinfo.get("relative_uri"):
                varinfo["relative_uri"] += f"&path={path}"
            else:
                varinfo["relative_uri"] = f"?path={path}"
        return True

    def __iter__(self):
        adict = self.value
        if isinstance(adict, dict):
            return iter(adict)
        else:
            return iter(tuple())

    def __len__(self):
        adict = self.value
        if isinstance(adict, dict):
            return len(adict)
        else:
            return 0

    def serialize(self):
        """Serialize data before persistent storage"""
        data = dict(self.metadata)
        data[CONTAINER_URI_NAME] = {
            name: var.data_proxy.serialize()
            for name, var in self.value.items()
            if var.data_proxy is not None
        }
        return data

    def deserialize(self, data: dict):
        """Deserialize data after loading from persistent storage"""
        variables = data.pop(CONTAINER_URI_NAME)
        value = dict()
        for name, data_proxy in variables.items():
            data_proxy = DataProxy.deserialize(data_proxy)
            if data_proxy is None:
                continue
            value[name] = self._create_variable(name, data_proxy)
        self._runtime_value = value
        self._runtime_metadata = data

    def dump(self):
        b = True
        for name, var in self.items():
            try:
                b &= var.dump()
            except Exception as e:
                raise RuntimeError(f"cannot serialize variable '{name}'") from e
        b &= super().dump()
        return b

    @property
    def container_has_persistent_value(self):
        return super()._has_persistent_value()

    def _has_persistent_value(self):
        if self.container_has_persistent_value:
            return all(v.has_persistent_value for v in self.values())
        else:
            return False

    @property
    def container_has_runtime_value(self):
        return super()._has_runtime_value()

    def _has_runtime_value(self):
        if self.container_has_runtime_value:
            return all(v.has_runtime_value for v in self.values())
        else:
            return False

    @property
    def container_has_value(self):
        return self.container_has_runtime_value or self.container_has_persistent_value

    def force_non_existing(self):
        super().force_non_existing()
        for v in self.values():
            v.force_non_existing()

    @property
    def variable_uhashes(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'variable_uhashes' is deprecated in favor of the function 'get_variable_uhashes'",
            DeprecationWarning,
        )
        return self.get_variable_uhashes()

    def get_variable_uhashes(self):
        return {name: var.uhash for name, var in self.items()}

    @property
    def variable_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'variable_values' is deprecated in favor of the function 'get_variable_values'",
            DeprecationWarning,
        )
        return self.get_variable_values()

    def get_variable_values(self):
        return {k: v.value for k, v in self.items() if not v.is_missing()}

    @property
    def variable_data_proxies(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'variable_data_proxies' is deprecated in favor of the function 'get_variable_data_proxies'",
            DeprecationWarning,
        )
        return self.get_variable_data_proxies()

    def get_variable_data_proxies(self):
        return {k: v.data_proxy for k, v in self.items()}

    @property
    def variable_uris(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'variable_uris' is deprecated in favor of the function 'get_variable_uris'",
            DeprecationWarning,
        )
        return self.get_variable_uris()

    def get_variable_uris(self):
        uris = dict()
        for k, v in self.items():
            proxy = v.data_proxy
            if proxy:
                uris[k] = proxy.uri
        return uris

    @property
    def variable_transfer_data(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'variable_transfer_data' is deprecated in favor of the function 'get_variable_transfer_data'",
            DeprecationWarning,
        )
        return self.get_variable_transfer_data()

    def get_variable_transfer_data(self):
        """Transfer data by variable or URI"""
        data = dict()
        for name, var in self.items():
            if var.has_persistent_value:
                data[name] = var.data_proxy.uri
            elif var.hashing_enabled:
                # Remove possible references to a uhashable
                data[name] = var.copy_without_references()
            else:
                data[name] = var.value
        return data

    @property
    def named_variable_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'named_variable_values' is deprecated in favor of the function 'get_named_variable_values'",
            DeprecationWarning,
        )
        return self.get_named_variable_values()

    def get_named_variable_values(self):
        return {
            k: v.value
            for k, v in self.items()
            if isinstance(k, str) and not v.is_missing()
        }

    @property
    def positional_variable_values(self):
        """DEPRECATED"""
        warnings.warn(
            "the property 'positional_variable_values' is deprecated in favor of the function 'get_positional_variable_values'",
            DeprecationWarning,
        )
        return self.get_positional_variable_values()

    def get_positional_variable_values(self):
        values = [self.MISSING_DATA] * self.__npositional_vars
        for i, var in self.items():
            if isinstance(i, int):
                values[i] = var.value
        return tuple(values)


def variable_from_transfer(data, varinfo=None) -> Variable:
    """Meant for task schedulers that pass data (see `VariableContainer.variable_transfer_data`)"""
    if isinstance(data, Variable):
        return data
    kw = {"varinfo": varinfo}
    if isinstance(data, DataProxy):
        kw["data_proxy"] = data
    elif isinstance(data, DataUri):
        kw["data_uri"] = data
    elif isinstance(data, (hashing.UniversalHash, hashing.HasUhash)):
        kw["pre_uhash"] = data
    else:
        kw["value"] = data
    return Variable(**kw)


def value_from_transfer(data, varinfo=None):
    """Meant for task schedulers that pass data (see VariableContainer.variable_transfer_*)"""
    if isinstance(data, Variable):
        return data.value
    kw = {"varinfo": varinfo}
    if isinstance(data, DataProxy):
        kw["data_proxy"] = data
    elif isinstance(data, DataUri):
        kw["data_uri"] = data
    elif isinstance(data, (hashing.UniversalHash, hashing.HasUhash)):
        kw["pre_uhash"] = data
    else:
        return data
    return Variable(**kw).value


class MutableVariableContainer(VariableContainer, MutableMapping):
    """An mutable mapping of variable identifiers (str or int) to variables (Variable)."""

    def __setitem__(self, key, value):
        self._set_item(key, value)

    def __delitem__(self, key):
        adict = self.value
        if isinstance(adict, dict):
            del self.value[key]

    def update_values(self, items):
        if isinstance(items, Mapping):
            items = items.items()
        for k, v in items:
            self[k].value = v


class MissingVariableError(RuntimeError):
    pass


class ReadOnlyVariableError(RuntimeError):
    pass


class ReadOnlyVariableContainerNamespace:
    """Expose getting variable values through attributes and indexing"""

    def __init__(self, container):
        self._container = container

    _RESERVED_VARIABLE_NAMES = None

    @classmethod
    def _reserved_variable_names(cls):
        if cls._RESERVED_VARIABLE_NAMES is None:
            cls._RESERVED_VARIABLE_NAMES = set(dir(cls)) | {"_container"}
        return cls._RESERVED_VARIABLE_NAMES

    def __setattr__(self, attrname, value):
        if attrname == "_container":
            super().__setattr__(attrname, value)
        else:
            self._get_variable(attrname)
            raise ReadOnlyVariableError(attrname)

    def __getattr__(self, attrname):
        return self[attrname]

    def __getitem__(self, key):
        return self._get_variable(key).value

    def __setitem__(self, key, value):
        raise ReadOnlyVariableError(key)

    def _get_variable(self, key):
        try:
            return self._container[key]
        except (KeyError, TypeError):
            raise MissingVariableError(key) from None


class VariableContainerNamespace(ReadOnlyVariableContainerNamespace):
    """Expose getting and setting variable values through attributes and indexing"""

    def __setattr__(self, attrname, value):
        if attrname == "_container":
            super().__setattr__(attrname, value)
        else:
            self[attrname] = value

    def __setitem__(self, key, value):
        self._get_variable(key).value = value


class VariableContainerMissingNamespace(ReadOnlyVariableContainerNamespace):
    """Expose missing variable values through attributes and indexing"""

    def __init__(self, container):
        self._container = container

    def __getitem__(self, key):
        return self._is_missing(key)

    def _is_missing(self, key):
        try:
            var = self._container[key]
        except (KeyError, TypeError):
            raise MissingVariableError(key) from None
        return var.is_missing()



import pytest
import logging
from ewokscore.graph import load_graph
from ewokscore.graph.schema import LATEST_VERSION

LATEST_VERSION = str(LATEST_VERSION)


def test_graph_version(caplog):
    # Update of the default version
    with caplog.at_level(logging.WARNING):
        assert_load({"graph": {"id": "test"}})

    # Update of the latest version
    assert_load({"graph": {"id": "test", "schema_version": LATEST_VERSION}})

    # Update method which raises an exception
    with pytest.raises(
        ValueError,
        match='Graph schema version "0.0" requires another library version: python3 -m pip install "ewokscore>=0.0,<0.0.1"',
    ):
        load_graph({"graph": {"id": "test", "schema_version": "0.0"}})

    # Update method which does not update the version
    with pytest.raises(
        AssertionError,
        match="graph conversion did not update the schema version",
    ):
        load_graph({"graph": {"id": "test", "schema_version": "0.1"}})

    # Correct update method
    assert_load({"graph": {"id": "test", "schema_version": "0.2"}})

    # Version does not exist
    with pytest.raises(
        ValueError,
        match='Graph schema version "99999.0" is either invalid or requires a newer library version: python3 -m pip install --upgrade ewokscore',
    ):
        load_graph({"graph": {"id": "test", "schema_version": "99999.0"}})


def assert_load(adict: dict):
    load_graph(adict).graph.graph["schema_version"] == LATEST_VERSION



import os
import sys
import pytest
from ewokscore.task import Task

WIN32 = sys.platform == "win32"


pyscript = r"""
if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--a", type=int, default=0)
    args = parser.parse_args()
    print("input a =", args.a)
    assert args.a == 10
"""


@pytest.mark.parametrize("shebang", [True, False])
@pytest.mark.parametrize("fail", [True, False])
def test_python_script_task(tmpdir, varinfo, shebang, fail):
    if WIN32 and shebang:
        pytest.skip("windows does not have shebangs")

    pyscriptname = tmpdir / "test.py"
    with open(pyscriptname, mode="w") as f:
        if shebang:
            f.write(f"#!{sys.executable}\n")
        f.writelines(pyscript)
    if not WIN32:
        os.chmod(pyscriptname, 0o755)

    if fail:
        a = 11
    else:
        a = 10

    task = Task.instantiate(
        "ScriptExecutorTask",
        inputs={
            "a": a,
            "_script": str(pyscriptname),
            "_capture_output": True,
            "_raise_on_error": False,
        },
        varinfo=varinfo,
    )
    task.execute()
    _assert_outputs(task, a, fail)


if WIN32:
    shellscript = r"""@echo off

set a=0

:initial
if "%1"=="" goto done
echo              %1
set aux=%1
if "%aux:~0,1%"=="-" (
   set varname=%aux:~1,250%
) else (
   set "%varname%=%1"
   set varname=
)
shift
goto initial
:done

echo input a = %a%
if %a%==10 (
    exit 0
) else (
    echo "failure" 1>&2
    exit 1
)
"""
else:
    shellscript = r"""a=0

while getopts u:a:f: flag
do
    case "${flag}" in
        a) a=${OPTARG};;
    esac
done

echo "input a = "$a
if [[ $a == "10" ]]; then
    exit 0
else
    echo "failure" 1>&2
    exit 1
fi
"""


@pytest.mark.parametrize("shebang", [True, False])
@pytest.mark.parametrize("fail", [True, False])
def test_shell_script_task(tmpdir, varinfo, shebang, fail):
    if WIN32:
        if shebang:
            pytest.skip("windows does not have shebangs")
        ext = ".bat"
    else:
        ext = ".sh"
    filename = tmpdir / f"test{ext}"
    with open(filename, mode="w") as f:
        if shebang:
            f.write("#!/bin/bash\n")
        f.writelines(shellscript)
    if not WIN32:
        os.chmod(filename, 0o755)

    if fail:
        a = 11
    else:
        a = 10

    task = Task.instantiate(
        "ScriptExecutorTask",
        inputs={
            "a": a,
            "_script": str(filename),
            "_capture_output": True,
            "_raise_on_error": False,
        },
        varinfo=varinfo,
    )
    task.execute()
    _assert_outputs(task, a, fail)


def test_command_task(tmpdir, varinfo):
    filename = tmpdir / "test.txt"
    with open(filename, mode="w"):
        pass

    task = Task.instantiate(
        "ScriptExecutorTask",
        inputs={
            "_script": f"dir {tmpdir}",
            "_capture_output": True,
            "_raise_on_error": False,
        },
        varinfo=varinfo,
    )
    task.execute()
    assert task.done
    assert "test.txt" in task.outputs.out


def _assert_outputs(task, a, fail):
    assert task.done
    out = "".join(task.outputs.out)
    assert f"input a = {a}" in out
    err = "".join(task.outputs.err)
    if fail:
        assert task.outputs.return_code != 0
        assert err
    else:
        assert task.outputs.return_code == 0
        assert not err



"""Initialize ewoks event fields
"""

import os
import getpass
import socket
from uuid import uuid4
from typing import Mapping, Optional, Union
import networkx

ExecInfoType = Union[Mapping, None]


def init_job(execinfo: ExecInfoType, **static_job_info) -> ExecInfoType:
    if execinfo is None:
        return None
    execinfo = dict(execinfo)
    set_environment(execinfo)
    job_id = execinfo.get("job_id")
    if job_id is None:
        execinfo["job_id"] = str(uuid4())
    execinfo.update(static_job_info)
    return execinfo


def init_workflow(
    execinfo: ExecInfoType,
    workflow: Union[networkx.DiGraph, str, None] = None,
    **static_workflow_info,
) -> ExecInfoType:
    if execinfo is None:
        return None
    execinfo = dict(execinfo)
    set_environment(execinfo)
    if workflow is None:
        default = (
            f"{execinfo['host_name']}-{execinfo['process_id']}-{execinfo['user_name']}"
        )
        execinfo["workflow_id"] = default
    elif isinstance(workflow, str):
        execinfo["workflow_id"] = workflow
    else:
        try:
            execinfo["workflow_id"] = str(workflow.graph["id"])
        except KeyError:
            raise ValueError("the graph needs an 'id' for execution events")
    execinfo.update(static_workflow_info)
    return execinfo


def init_node(
    execinfo: ExecInfoType, node_id: Optional[str], task_id: Optional[str]
) -> ExecInfoType:
    if execinfo is None:
        return None
    execinfo = dict(execinfo)
    set_environment(execinfo)
    if node_id:
        execinfo["node_id"] = node_id
    if task_id:
        execinfo["task_id"] = task_id
    return execinfo


def set_environment(execinfo: ExecInfoType) -> ExecInfoType:
    if execinfo is None:
        return None
    execinfo["host_name"] = socket.gethostname()
    execinfo["process_id"] = os.getpid()
    execinfo["user_name"] = getpass.getuser()



import gc
import pytest
import json
from glob import glob
from pathlib import Path
from ewokscore.task import Task
from ewokscore.task import TaskInputError
from .examples.tasks.sumtask import SumTask


def find_files(tmpdir: Path, extension):
    return glob(str(tmpdir / "**" / "*" + extension), recursive=True)


def expected_task_output_storage(task):
    expected = [var.serialize() for var in task.output_variables.values()]
    expected.append(task.output_variables.serialize())
    return expected


def assert_storage(tmpdir, expected):
    lst = []
    for filename in find_files(tmpdir, ".json"):
        with open(filename, "r") as fileobj:
            lst.append(json.load(fileobj))
    for v in lst:
        if isinstance(v, dict):
            v.pop("__traceback__", None)
    assert len(lst) == len(expected)
    for v in expected:
        lst.pop(lst.index(v))
    assert not lst, "Unexpected data saved"


def test_no_public_reserved_names():
    assert not [s for s in Task._reserved_variable_names() if not s.startswith("_")]


def test_task_missing_input():
    with pytest.raises(TaskInputError):
        SumTask()


def test_task_readonly_input():
    task = SumTask(inputs={"a": 10})
    with pytest.raises(RuntimeError):
        task.inputs.a = 10


def test_task_optional_input(tmpdir, varinfo):
    task = SumTask(inputs={"a": 10}, varinfo=varinfo)
    assert not task.done
    task.execute()
    assert task.done
    assert task.outputs.result == 10
    expected = expected_task_output_storage(task)
    assert_storage(tmpdir, expected)


def test_task_done(varinfo):
    task = SumTask(inputs={"a": 10}, varinfo=varinfo)
    assert not task.done
    task.execute()
    assert task.done

    task = SumTask(inputs={"a": 10}, varinfo=varinfo)
    assert task.done

    task = SumTask(inputs={"a": 10})
    assert not task.done
    task.execute()
    assert task.done

    task = SumTask(inputs={"a": 10})
    assert not task.done


def test_task_uhash(varinfo):
    task = SumTask(inputs={"a": 10}, varinfo=varinfo)
    uhash = task.uhash
    assert task.uhash == task.output_variables.uhash
    assert task.uhash != task.input_variables.uhash

    task.input_variables["a"].value += 1
    assert task.uhash != uhash
    assert task.uhash == task.output_variables.uhash
    assert task.uhash != task.input_variables.uhash


def test_task_storage(tmpdir, varinfo):
    task = SumTask(inputs={"a": 10, "b": 2}, varinfo=varinfo)
    assert not task.done
    task.execute()
    assert task.done
    assert task.outputs.result == 12
    expected = expected_task_output_storage(task)
    assert_storage(tmpdir, expected)

    task = SumTask(inputs={"a": 10, "b": 2}, varinfo=varinfo)
    assert task.done
    assert task.outputs.result == 12
    assert_storage(tmpdir, expected)

    task = SumTask({"a": 2, "b": 10}, varinfo=varinfo)
    assert not task.done
    task.execute()
    assert task.done
    assert task.outputs.result == 12
    expected += expected_task_output_storage(task)
    assert_storage(tmpdir, expected)

    task = SumTask({"a": task.output_variables["result"], "b": 0}, varinfo=varinfo)
    assert not task.done
    task.execute()
    assert task.done
    assert task.outputs.result == 12
    expected += expected_task_output_storage(task)
    assert_storage(tmpdir, expected)

    task = SumTask(
        {"a": 1, "b": task.output_variables["result"].data_proxy}, varinfo=varinfo
    )
    assert not task.done
    task.execute()
    assert task.done
    assert task.outputs.result == 13
    expected += expected_task_output_storage(task)
    assert_storage(tmpdir, expected)

    task = SumTask(
        {"a": 1, "b": task.output_variables["result"].data_proxy.uri}, varinfo=varinfo
    )
    assert not task.done
    task.execute()
    assert task.done
    assert task.outputs.result == 14
    expected += expected_task_output_storage(task)
    assert_storage(tmpdir, expected)


def test_task_required_positional_inputs():
    class MyTask(Task, n_required_positional_inputs=1):
        pass

    with pytest.raises(TaskInputError):
        MyTask()


def test_task_cleanup_references():
    class MyTask(Task, input_names=["mylist"], output_names=["mylist"]):
        def run(self):
            self.outputs.mylist = self.inputs.mylist + [len(self.inputs.mylist)]

    obj = [0, 1, 2]
    nref_start = len(gc.get_referrers(obj))
    task1 = MyTask(inputs={"mylist": obj})
    task2 = MyTask(inputs=task1.output_variables)
    task1.execute()
    task2.execute()
    assert len(gc.get_referrers(obj)) > nref_start

    uhash1 = task1.uhash
    uhashes1 = task1.get_output_uhashes()
    uhash2 = task2.uhash
    uhashes2 = task2.get_output_uhashes()

    task1.cleanup_references()

    while gc.collect():
        pass
    assert len(gc.get_referrers(obj)) == nref_start

    assert uhash1 == task1.uhash
    assert uhashes1 == task1.get_output_uhashes()
    assert uhash2 == task2.uhash
    assert uhashes2 == task2.get_output_uhashes()



import sys
from enum import Enum
import logging

_logger = logging.getLogger(__name__)


class BaseProgress:
    """
    Interface to define a Task progress.
    """

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, progress):
        self._progress = progress
        self.update()

    def reset(self):
        raise NotImplementedError("Base class")

    def update(self):
        """
        Update is call when the progress evolves
        :return:
        """
        raise NotImplementedError("Base class")


class BasePercentageProgress(BaseProgress):
    """
    Define a progress in [0, 100] %
    """

    def __init__(self):
        super().__init__()
        self._progress = 0
        self._lastUpdate = None

    @BaseProgress.progress.setter
    def progress(self, progress):
        if not isinstance(progress, (int, float)):
            raise TypeError("Progress is expected to be an int or a float")
        if not (0 <= progress <= 100):
            _logger.warning("progress is expected to be in [0, 100]. Clip it")
            progress = min(max(0, progress), 100)
        BaseProgress.progress.fset(self, int(progress))

    def reset(self):
        self._lastUpdate = None
        self._progress = 0
        self.update()

    def update(self):
        if self._progress != self._lastUpdate:
            self._lastUpdate = self._progress
            self._update()

    def _update(self):
        raise NotImplementedError("Base class")


class _TextAdvancement(Enum):
    step_1 = "\\"
    step_2 = "-"
    step_3 = "/"
    step_4 = "|"

    @staticmethod
    def getNextStep(step):
        if step is _TextAdvancement.step_1:
            return _TextAdvancement.step_2
        elif step is _TextAdvancement.step_2:
            return _TextAdvancement.step_3
        elif step is _TextAdvancement.step_3:
            return _TextAdvancement.step_4
        else:
            return _TextAdvancement.step_1

    @staticmethod
    def getStep(value):
        if value % 4 == 0:
            return _TextAdvancement.step_4
        elif value % 3 == 0:
            return _TextAdvancement.step_3
        elif value % 2 == 0:
            return _TextAdvancement.step_2
        else:
            return _TextAdvancement.step_1


class TextProgress(BasePercentageProgress):
    """Print advancement in sys.stdout"""

    def __init__(self, name, char_length=20):
        super().__init__()
        self._name = name
        self._char_length = char_length

    @property
    def char_length(self):
        """how many character we want to use to represent the advancement"""
        return self._char_length

    @char_length.setter
    def char_length(self, length: int):
        if not isinstance(length, int):
            raise TypeError("length is expected to be an int")
        self._char_length = length

    @property
    def name(self) -> str:
        return self._name

    def _update(self):
        progress = self.progress
        if progress is None:
            progress = 0
        block = int(round(self.char_length * progress / 100))
        msg = "\r{0}: [{1}] {2}%".format(
            self.name,
            "#" * block + "-" * (self.char_length - block),
            round(progress, 2),
        )
        if progress >= 100:
            msg += " DONE\r\n"
        sys.stdout.write(msg)
        sys.stdout.flush()



import gc
import pytest
from contextlib import contextmanager
from ewokscore.missing_data import MISSING_DATA
from ewokscore.variable import Variable, VariableContainer
from ewokscore.variable import MutableVariableContainer

VALUES = [None, True, 10, "string", 10.1, [1, 2, 3], {"1": 1, "2": {"2": [10, 20]}}]


def test_variable_no_value(varinfo):
    var = Variable(varinfo=varinfo)
    assert_variable_no_value(var)


def test_variable_container_no_value(varinfo):
    var = MutableVariableContainer(varinfo=varinfo)
    assert_variable_no_value(var)


def test_variable_container_wrong_value():
    with pytest.raises(TypeError):
        MutableVariableContainer(value=[1, 2])


def assert_variable_no_value(var):
    assert not var.has_runtime_value
    assert not var.has_persistent_value
    assert not var.value
    assert not var.data_uri
    var.dump()
    var.load()
    assert not var.has_runtime_value
    assert not var.has_persistent_value
    assert not var.value
    assert not var.data_uri
    assert var.value is var.MISSING_DATA
    assert var.value == var.MISSING_DATA


@pytest.mark.parametrize("value", VALUES)
def test_variable_references(value, varinfo):
    var = Variable(value, varinfo=varinfo)
    references = {
        "same_instance": var,
        "same_value": Variable(value, varinfo=varinfo),
        "ref_uhashable": Variable(pre_uhash=var, varinfo=varinfo),
        "ref_uhash": Variable(pre_uhash=var.uhash, varinfo=varinfo),
        "ref_proxy": Variable(data_proxy=var.data_proxy),
        "ref_uri": Variable(data_uri=var.data_uri),
    }
    assert_variable(var, **references)


def test_variable_container_references(varinfo):
    value = {f"var{i}": value for i, value in enumerate(VALUES, 1)}
    var = MutableVariableContainer(value=value, varinfo=varinfo)
    references = {
        "same_instance": var,
        "same_value": MutableVariableContainer(value=value, varinfo=varinfo),
        "ref_uhashable": MutableVariableContainer(pre_uhash=var, varinfo=varinfo),
        "ref_uhash": MutableVariableContainer(pre_uhash=var.uhash, varinfo=varinfo),
        "ref_proxy": MutableVariableContainer(data_proxy=var.data_proxy),
        "ref_uri": MutableVariableContainer(data_uri=var.data_uri),
        "ref_same_variables": MutableVariableContainer(value=var, varinfo=varinfo),
    }
    assert_variable(var, **references)


def assert_variable(var, **references):
    assert_variable_compare(var, **references)
    assert_variable_uhash(var, **references)
    assert_variable_uri(var, **references)
    assert_variable_data_proxy(var, **references)
    assert_variable_value(var, **references)
    assert_variable_value_status(var, **references)
    var.dump()
    assert_variable_compare(var, **references)
    assert_variable_uhash(var, **references)
    assert_variable_uri(var, **references)
    assert_variable_data_proxy(var, **references)
    assert_persisted_variable_value(var, **references)
    assert_persisted_variable_value_status(var, **references)


def assert_variable_uhash(var, **references):
    for name, var_ref in references.items():
        assert var.uhash == var_ref.uhash, name
    with change_value(var):
        for name, var_ref in references.items():
            if name in ("same_instance", "ref_uhashable", "ref_same_variables"):
                assert var.uhash == var_ref.uhash, name
            else:
                assert var.uhash != var_ref.uhash, name


def assert_variable_compare(var, **references):
    for name, var_ref in references.items():
        try:
            assert var == var_ref, name
        except AssertionError:
            var == var_ref
            raise
    with change_value(var):
        for name, var_ref in references.items():
            if name in ("same_instance", "ref_uhashable", "ref_same_variables"):
                assert var == var_ref, name
            else:
                assert var != var_ref, name


def assert_variable_data_proxy(var, **references):
    for name, var_ref in references.items():
        assert var.data_proxy == var_ref.data_proxy, name
    with change_value(var):
        for name, var_ref in references.items():
            if name in (
                "same_instance",
                "ref_uhashable",
                "ref_same_variables",
                "ref_proxy",
            ):
                assert var.data_proxy == var_ref.data_proxy, name
            else:
                assert var.data_proxy != var_ref.data_proxy, name


def assert_variable_uri(var, **references):
    for name, var_ref in references.items():
        assert var.data_uri == var_ref.data_uri, name
    with change_value(var):
        for name, var_ref in references.items():
            if name in (
                "same_instance",
                "ref_uhashable",
                "ref_same_variables",
                "ref_proxy",
            ):
                assert var.data_uri == var_ref.data_uri, name
            else:
                assert var.data_uri != var_ref.data_uri, name


def assert_variable_value(var, **references):
    # Side effect: variable will have runtime values (lazy loading in the `value` property)
    for name, var_ref in references.items():
        if name in ("same_instance", "same_value", "ref_same_variables"):
            assert var.value == var_ref.value, name
        else:
            assert var.value != var_ref.value, name
    with change_value(var):
        for name, var_ref in references.items():
            if name in ("same_instance", "ref_same_variables"):
                assert var.value == var_ref.value, name
            else:
                assert var.value != var_ref.value, name


def assert_persisted_variable_value(var, **references):
    # Assumption: values have been loaded
    for name, var_ref in references.items():
        assert var.value == var_ref.value, name
    with change_value(var):
        for name, var_ref in references.items():
            if name in ("same_instance", "ref_same_variables"):
                assert var.value == var_ref.value, name
            else:
                assert var.value != var_ref.value, name


def assert_variable_value_status(var, **references):
    # Assumption: values have not been loaded
    is_container = isinstance(var, MutableVariableContainer)

    for name, var_ref in references.items():
        if name in ("same_instance", "same_value", "ref_same_variables"):
            if is_container:
                assert not var_ref.container_has_persistent_value, name
                assert var_ref.container_has_runtime_value, name
            assert not var_ref.has_persistent_value, name
            assert var_ref.has_runtime_value, name
        else:
            if is_container:
                assert not var_ref.container_has_persistent_value, name
                assert not var_ref.container_has_runtime_value, name
            assert not var_ref.has_persistent_value, name
            assert not var_ref.has_runtime_value, name


def assert_persisted_variable_value_status(var, **references):
    # Assumption: values have been loaded
    # Side effect: container variables will have runtime values (explicit loading)
    is_container = isinstance(var, MutableVariableContainer)

    for name, var_ref in references.items():
        if is_container:
            assert var_ref.container_has_persistent_value, name
            assert var_ref.container_has_runtime_value, name
        else:
            assert var_ref.has_persistent_value, name
            assert var_ref.has_runtime_value, name
        if is_container:
            if name in ("same_instance", "same_value", "ref_same_variables"):
                assert var_ref.has_runtime_value, name
            else:
                assert not var_ref.has_runtime_value, name
                for sub_var in var_ref.value.values():
                    sub_var.load()
                assert var_ref.has_runtime_value, name


@contextmanager
def change_value(var):
    if isinstance(var, MutableVariableContainer):
        name = next(iter(var))
        keep = var[name].value
        var[name].value = 9999
    else:
        keep = var.value
        var.value = 9999
    try:
        yield
    finally:
        if isinstance(var, MutableVariableContainer):
            var[name].value = keep
        else:
            var.value = keep


def test_variable_nonce(varinfo):
    v1 = Variable(9999, varinfo=varinfo)
    v2 = Variable(value=9999, instance_nonce=1, varinfo=varinfo)
    assert v1.uhash != v2.uhash
    assert v1 != v2
    assert v1.value == v2.value
    v2 = Variable(pre_uhash=v1, instance_nonce=1, varinfo=varinfo)
    assert v1.uhash != v2.uhash
    assert v1 != v2
    assert v1.value != v2.value
    v2 = Variable(pre_uhash=v1.uhash, instance_nonce=1, varinfo=varinfo)
    assert v1.uhash != v2.uhash
    assert v1 != v2
    assert v1.value != v2.value


@pytest.mark.parametrize("scheme", ("json", "nexus"))
@pytest.mark.parametrize("root_uri_type", ("directory", "filename", "path_in_file"))
def test_variable_container_metadata(scheme, root_uri_type, tmpdir):
    if scheme == "nexus":
        extension = ".nx"
    else:
        extension = ".json"

    if root_uri_type == "path_in_file":
        root_uri = tmpdir / f"dataset_name{extension}::/scan_name"
    elif root_uri_type == "filename":
        root_uri = tmpdir / f"dataset_name{extension}"
    elif root_uri_type == "directory":
        root_uri = tmpdir
    varinfo = {"root_uri": str(root_uri), "scheme": scheme}

    values = {f"var{i}": i for i in range(3)}
    container = MutableVariableContainer(value=values, varinfo=varinfo)
    container.metadata["myvalue"] = 999
    container["var1"].metadata["myvalue"] = 888
    container.dump()

    # Common ways to receive data
    ref_uri = MutableVariableContainer(data_uri=container.data_uri)
    ref_transfer_data = MutableVariableContainer(
        value=container.get_variable_transfer_data(), varinfo={"has_data_proxy": True}
    )
    assert ref_uri.uhash == container.uhash
    assert ref_transfer_data.uhash == container.uhash

    # Check data
    assert container.get_variable_values() == ref_transfer_data.get_variable_values()
    assert (
        container.get_variable_transfer_data()
        == ref_transfer_data.get_variable_transfer_data()
    )
    assert container.get_variable_values() == ref_uri.get_variable_values()
    assert (
        container.get_variable_transfer_data() == ref_uri.get_variable_transfer_data()
    )

    # Check metadata
    if scheme == "nexus":
        if root_uri_type == "path_in_file":
            assert ref_uri.metadata["@NX_class"] == "NXcollection"
        else:
            assert ref_uri.metadata["@NX_class"] == "NXprocess"

    assert ref_uri.metadata["myvalue"] == 999

    if scheme == "nexus":
        assert ref_uri["var1"].metadata["@NX_class"] == "NXcollection"
    assert ref_uri["var1"].metadata["myvalue"] == 888


def test_variable_cleanup_references():
    obj = [0, 1, 2]
    nref_start = len(gc.get_referrers(obj))
    var1 = Variable(obj)
    var2 = Variable(pre_uhash=var1)
    uhash = var1.uhash
    assert uhash == var2.uhash
    assert len(gc.get_referrers(obj)) > nref_start

    del var1
    while gc.collect():
        pass
    assert len(gc.get_referrers(obj)) > nref_start

    var2.cleanup_references()
    while gc.collect():
        pass
    assert len(gc.get_referrers(obj)) == nref_start

    assert uhash == var2.uhash


def test_variable_container_cleanup_references():
    obj = [0, 1, 2]
    nref_start = len(gc.get_referrers(obj))
    var1 = MutableVariableContainer({"myvar": obj})
    var2 = MutableVariableContainer(pre_uhash=var1)
    uhash = var1.uhash
    assert uhash == var2.uhash

    del var1
    while gc.collect():
        pass
    assert len(gc.get_referrers(obj)) > nref_start

    var2.cleanup_references()
    while gc.collect():
        pass
    assert len(gc.get_referrers(obj)) == nref_start

    assert uhash == var2.uhash


def test_variable_fixed_uhash():
    class MyClass:
        pass

    var = Variable(value=MyClass, varinfo={"enable_hashing": True})
    with pytest.raises(TypeError):
        assert var.uhash

    var = Variable(
        value=MyClass, varinfo={"enable_hashing": True, "uhash_data": "some data"}
    )
    assert var.uhash

    var1 = Variable(value=10, varinfo={"enable_hashing": True, "uhash_data": None})
    var2 = Variable(
        value=10, varinfo={"enable_hashing": True, "uhash_data": "some data"}
    )
    var3 = Variable(
        value=20, varinfo={"enable_hashing": True, "uhash_data": "some data"}
    )
    assert var1.uhash
    assert var2.uhash
    assert var3.uhash
    assert var1.uhash != var2.uhash
    assert var2.uhash == var3.uhash


def test_variable_uri(tmpdir):
    var1 = Variable(value=10, varinfo={"root_uri": str(tmpdir)})
    var1.dump()

    var2 = Variable(data_uri=str(var1.data_uri))
    assert var1.value == 10
    assert var1.value == var2.value
    assert var1 == var2


def test_variable_container_uri(tmpdir):
    var1 = VariableContainer(value={"a": 1, "b": 2}, varinfo={"root_uri": str(tmpdir)})
    var1.dump()

    var2 = VariableContainer(data_uri=str(var1.data_uri))
    assert var1.get_variable_values() == {"a": 1, "b": 2}
    assert var1.value == var2.value
    assert var1 == var2


@pytest.mark.parametrize("scheme", ("json", "nexus"))
def test_variable_none_dump(scheme, tmpdir):
    var = Variable(value=None, varinfo={"root_uri": str(tmpdir), "scheme": scheme})
    var.dump()
    var.load()


def test_variable_container_values():
    c = MutableVariableContainer(
        value={"a": None, "b": MISSING_DATA, 2: "Two", 3: MISSING_DATA}
    )
    assert not c["a"].is_missing()
    assert c["b"].is_missing()
    assert c.get_variable_values() == {"a": None, 2: "Two"}
    assert c.get_named_variable_values() == {"a": None}
    assert c.get_positional_variable_values() == (
        MISSING_DATA,
        MISSING_DATA,
        "Two",
        MISSING_DATA,
    )

    c["a"] = MISSING_DATA
    c["b"] = None
    c[1] = "One"
    c[2] = MISSING_DATA
    assert c["a"].is_missing()
    assert not c["b"].is_missing()
    assert c.get_variable_values() == {"b": None, 1: "One"}
    assert c.get_named_variable_values() == {"b": None}
    assert c.get_positional_variable_values() == (
        MISSING_DATA,
        "One",
        MISSING_DATA,
        MISSING_DATA,
    )



from typing import Optional, Dict, List, Union, Any
from collections import Counter
import networkx

from ...node import NodeIdType
from ...task import Task
from ...inittask import instantiate_task as _instantiate_task
from ...inittask import add_dynamic_inputs
from .. import analysis
from .. import graph_io
from ... import events


def instantiate_task(graph: networkx.DiGraph, node_id: NodeIdType, **kw) -> Task:
    """Named arguments are dynamic input and Variable config.
    Default input from the persistent representation are added internally.
    """
    # Dynamic input has priority over default input
    nodeattrs = graph.nodes[node_id]
    return _instantiate_task(node_id, nodeattrs, **kw)


def instantiate_task_static(
    graph: networkx.DiGraph,
    node_id: NodeIdType,
    tasks: Optional[Dict[Task, int]] = None,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    evict_result_counter: Optional[Dict[NodeIdType, int]] = None,
) -> Task:
    """Instantiate destination task while no access to the dynamic inputs.
    Side effect: `tasks` will contain all predecessors.
    """
    if analysis.graph_is_cyclic(graph):
        raise RuntimeError("cannot execute cyclic graphs with ewokscore")
    if tasks is None:
        tasks = dict()
    if evict_result_counter is None:
        evict_result_counter = dict()
    # Input from previous tasks (instantiate them if needed)
    dynamic_inputs = dict()
    for source_id in analysis.node_predecessors(graph, node_id):
        source_task = tasks.get(source_id, None)
        if source_task is None:
            source_task = instantiate_task_static(
                graph,
                source_id,
                tasks=tasks,
                varinfo=varinfo,
                execinfo=execinfo,
                evict_result_counter=evict_result_counter,
            )
        link_attrs = graph[source_id][node_id]
        add_dynamic_inputs(
            dynamic_inputs,
            link_attrs,
            source_task.output_variables,
            source_id=source_id,
            target_id=node_id,
        )
        # Evict intermediate results
        if evict_result_counter:
            evict_result_counter[source_id] -= 1
            if evict_result_counter[source_id] == 0:
                tasks.pop(source_id)
    # Instantiate the requested task
    target_task = instantiate_task(
        graph, node_id, inputs=dynamic_inputs, varinfo=varinfo, execinfo=execinfo
    )
    tasks[node_id] = target_task
    return target_task


def successor_counter(graph: networkx.DiGraph) -> Dict[NodeIdType, int]:
    nsuccessor = Counter()
    for edge in graph.edges:
        nsuccessor[edge[0]] += 1
    return nsuccessor


def execute_graph(
    graph: networkx.DiGraph,
    varinfo: Optional[dict] = None,
    execinfo: Optional[dict] = None,
    raise_on_error: Optional[bool] = True,
    outputs: Optional[List[dict]] = None,
    merge_outputs: Optional[bool] = True,
    output_tasks: Optional[bool] = False,
) -> Union[Dict[NodeIdType, Task], Dict[str, Any]]:
    """Sequential execution of DAGs.

    When `output_tasks` is `True` the arguments `outputs` and `merge_outputs`
    are ignored and instead of returning output values, it returns `Task` instances.
    This was introduced for testing.
    """
    with events.workflow_context(execinfo, workflow=graph) as execinfo:
        if analysis.graph_is_cyclic(graph):
            raise RuntimeError("cannot execute cyclic graphs")
        if analysis.graph_has_conditional_links(graph):
            raise RuntimeError("cannot execute graphs with conditional links")

        # Pepare containers for local state
        tasks = dict()
        if output_tasks:
            output_values = None
            evict_result_counter = None
        else:
            outputs = graph_io.parse_outputs(graph, outputs)
            output_values = dict()
            evict_result_counter = successor_counter(graph)

        # Execute in topological order
        for node_id in analysis.topological_sort(graph):
            task = instantiate_task_static(
                graph,
                node_id,
                tasks=tasks,
                varinfo=varinfo,
                execinfo=execinfo,
                evict_result_counter=evict_result_counter,
            )
            task.execute(
                raise_on_error=raise_on_error,
                cleanup_references=evict_result_counter is not None,
            )
            if execinfo:
                execinfo.setdefault("exception", task.exception)
            if not output_tasks:
                graph_io.add_output_values(
                    output_values, node_id, task, outputs, merge_outputs=merge_outputs
                )

        # Return results or Task instances
        if output_tasks:
            return tasks
        else:
            return output_values
